from __future__ import annotations

import pygame
from pygame import surface

from ..colors import LIGHT_GRAY, WHITE
from ..font_utils import load_font, render_text_surface
from ..localization import get_font_settings
from ..localization import translate as tr
from ..render_constants import GAMEPLAY_FONT_SIZE


def _wrap_long_segment(
    segment: str, font: pygame.font.Font, max_width: int
) -> list[str]:
    lines: list[str] = []
    current = ""
    for char in segment:
        candidate = current + char
        if font.size(candidate)[0] <= max_width or not current:
            current = candidate
        else:
            lines.append(current)
            current = char
    if current:
        lines.append(current)
    return lines


def wrap_text(text: str, font: pygame.font.Font, max_width: int) -> list[str]:
    if max_width <= 0:
        return [text]
    paragraphs = text.splitlines() or [text]
    lines: list[str] = []
    for paragraph in paragraphs:
        if not paragraph:
            lines.append("")
            continue
        words = paragraph.split(" ")
        if len(words) == 1:
            lines.extend(_wrap_long_segment(paragraph, font, max_width))
            continue
        current = ""
        for word in words:
            candidate = f"{current} {word}".strip() if current else word
            if font.size(candidate)[0] <= max_width:
                current = candidate
                continue
            if current:
                lines.append(current)
            if font.size(word)[0] <= max_width:
                current = word
            else:
                lines.extend(_wrap_long_segment(word, font, max_width))
                current = ""
        if current:
            lines.append(current)
    return lines


def blit_text_wrapped(
    target: surface.Surface,
    text: str,
    font: pygame.font.Font,
    color: tuple[int, int, int],
    topleft: tuple[int, int],
    max_width: int,
    *,
    line_height_scale: float = 1.0,
) -> None:
    """Render text with simple wrapping constrained to max_width."""

    x, y = topleft
    line_height = int(round(font.get_linesize() * line_height_scale))
    for line in wrap_text(text, font, max_width):
        if not line:
            y += line_height
            continue
        rendered = render_text_surface(
            font, line, color, line_height_scale=line_height_scale
        )
        target.blit(rendered, (x, y))
        y += line_height


def blit_message(
    screen: surface.Surface,
    text: str,
    size: int,
    color: tuple[int, int, int],
    position: tuple[int, int],
) -> None:
    try:
        font_settings = get_font_settings()
        scaled_size = font_settings.scaled_size(size)
        font = load_font(font_settings.resource, scaled_size)
        text_surface = render_text_surface(
            font, text, color, line_height_scale=font_settings.line_height_scale
        )
        text_rect = text_surface.get_rect(center=position)

        # Add a semi-transparent background rectangle for better visibility
        bg_padding = 15
        bg_rect = text_rect.inflate(bg_padding * 2, bg_padding * 2)
        bg_surface = pygame.Surface((bg_rect.width, bg_rect.height), pygame.SRCALPHA)
        bg_surface.fill((0, 0, 0, 180))
        screen.blit(bg_surface, bg_rect.topleft)

        screen.blit(text_surface, text_rect)
    except pygame.error as e:
        print(f"Error rendering font or surface: {e}")


def blit_message_wrapped(
    screen: surface.Surface,
    text: str,
    size: int,
    color: tuple[int, int, int],
    position: tuple[int, int],
    *,
    max_width: int,
    line_spacing: int = 2,
) -> None:
    try:
        font_settings = get_font_settings()
        font = load_font(font_settings.resource, font_settings.scaled_size(size))
        line_height_scale = font_settings.line_height_scale
        lines = wrap_text(text, font, max_width)
        if not lines:
            return
        rendered = [
            render_text_surface(font, line, color, line_height_scale=line_height_scale)
            for line in lines
        ]
        max_line_width = max(text_surface.get_width() for text_surface in rendered)
        line_height = int(round(font.get_linesize() * line_height_scale))
        total_height = line_height * len(rendered) + line_spacing * (len(rendered) - 1)

        center_x, center_y = position
        top = center_y - total_height // 2

        bg_padding = 15
        bg_width = max_line_width + bg_padding * 2
        bg_height = total_height + bg_padding * 2
        bg_rect = pygame.Rect(0, 0, bg_width, bg_height)
        bg_rect.center = (center_x, center_y)
        bg_surface = pygame.Surface((bg_rect.width, bg_rect.height), pygame.SRCALPHA)
        bg_surface.fill((0, 0, 0, 180))
        screen.blit(bg_surface, bg_rect.topleft)

        y = top
        for text_surface in rendered:
            text_rect = text_surface.get_rect(centerx=center_x, y=y)
            screen.blit(text_surface, text_rect)
            y += line_height + line_spacing
    except pygame.error as e:
        print(f"Error rendering font or surface: {e}")


def draw_pause_overlay(
    screen: pygame.Surface,
    *,
    menu_labels: list[str] | None = None,
    selected_index: int = 0,
) -> list[pygame.Rect]:
    screen_width, screen_height = screen.get_size()
    overlay = pygame.Surface((screen_width, screen_height), pygame.SRCALPHA)
    overlay.fill((0, 0, 0, 150))
    pause_radius = 34
    cx = screen_width // 2
    cy = screen_height // 2 - 20
    pygame.draw.circle(
        overlay,
        LIGHT_GRAY,
        (cx, cy),
        pause_radius,
        width=3,
    )
    bar_width = 6
    bar_height = 22
    gap = 8
    pygame.draw.rect(
        overlay,
        LIGHT_GRAY,
        (cx - gap - bar_width, cy - bar_height // 2, bar_width, bar_height),
    )
    pygame.draw.rect(
        overlay,
        LIGHT_GRAY,
        (cx + gap, cy - bar_height // 2, bar_width, bar_height),
    )
    screen.blit(overlay, (0, 0))
    blit_message(
        screen,
        tr("hud.paused"),
        GAMEPLAY_FONT_SIZE,
        WHITE,
        (screen_width // 2, cy - pause_radius - 14),
    )
    option_rects: list[pygame.Rect] = []
    if menu_labels:
        font_settings = get_font_settings()
        menu_font = load_font(font_settings.resource, font_settings.scaled_size(11))
        line_height = int(
            round(menu_font.get_linesize() * font_settings.line_height_scale)
        )
        row_height = line_height + 6
        menu_width = max(140, int(screen_width * 0.34))
        menu_top = cy + pause_radius + 10
        highlight_color = (70, 70, 70)
        for idx, label in enumerate(menu_labels):
            rect = pygame.Rect(
                screen_width // 2 - menu_width // 2,
                menu_top + idx * row_height,
                menu_width,
                row_height,
            )
            if idx == selected_index:
                pygame.draw.rect(screen, highlight_color, rect)
            text_surface = render_text_surface(
                menu_font,
                label,
                WHITE,
                line_height_scale=font_settings.line_height_scale,
            )
            text_rect = text_surface.get_rect(center=rect.center)
            screen.blit(text_surface, text_rect)
            option_rects.append(rect)
    return option_rects



"""Gameplay helpers and logic utilities."""

# ruff: noqa: F401

from .ambient import sync_ambient_palette_with_flashlights
from .footprints import get_shrunk_sprite, update_footprints
from .entity_interactions import check_interactions
from .layout import MapGenerationError, generate_level_from_blueprint
from .entity_updates import process_player_input, update_entities
from .spawn import (
    maintain_waiting_car_supply,
    nearest_waiting_car,
    place_buddies,
    place_empty_fuel_can,
    place_flashlights,
    place_fuel_can,
    place_fuel_station,
    place_new_car,
    place_shoes,
    setup_player_and_cars,
    spawn_initial_patrol_bots,
    spawn_spiky_plants,
    spawn_exterior_zombie,
    spawn_initial_zombies,
    spawn_survivors,
    spawn_waiting_car,
    spawn_weighted_zombie,
)
from .state import (
    carbonize_outdoor_zombies,
    initialize_game_state,
    schedule_timed_message,
    update_endurance_timer,
)
from .survivors import (
    add_survivor_message,
    apply_passenger_speed_penalty,
    calculate_car_speed_for_passengers,
    cleanup_survivor_messages,
    drop_survivors_from_car,
    handle_survivor_zombie_collisions,
    increase_survivor_capacity,
    random_survivor_conversion_line,
    respawn_buddies_near_player,
    update_survivors,
)
from .utils import (
    find_exterior_spawn_position,
    find_interior_spawn_positions,
    find_nearby_offscreen_spawn_position,
    rect_visible_on_screen,
)

__all__ = [
    "MapGenerationError",
    "generate_level_from_blueprint",
    "place_new_car",
    "place_fuel_can",
    "place_empty_fuel_can",
    "place_fuel_station",
    "place_flashlights",
    "place_shoes",
    "place_buddies",
    "find_interior_spawn_positions",
    "find_nearby_offscreen_spawn_position",
    "find_exterior_spawn_position",
    "spawn_survivors",
    "spawn_exterior_zombie",
    "spawn_weighted_zombie",
    "update_survivors",
    "nearest_waiting_car",
    "calculate_car_speed_for_passengers",
    "apply_passenger_speed_penalty",
    "increase_survivor_capacity",
    "spawn_waiting_car",
    "maintain_waiting_car_supply",
    "add_survivor_message",
    "random_survivor_conversion_line",
    "cleanup_survivor_messages",
    "drop_survivors_from_car",
    "handle_survivor_zombie_collisions",
    "respawn_buddies_near_player",
    "get_shrunk_sprite",
    "update_footprints",
    "initialize_game_state",
    "schedule_timed_message",
    "setup_player_and_cars",
    "spawn_initial_patrol_bots",
    "spawn_spiky_plants",
    "spawn_initial_zombies",
    "update_endurance_timer",
    "carbonize_outdoor_zombies",
    "process_player_input",
    "update_entities",
    "check_interactions",
    "sync_ambient_palette_with_flashlights",
]



"""Stage definitions and defaults."""

from __future__ import annotations

from .entities_constants import ZOMBIE_DECAY_DURATION_FRAMES
from .gameplay_constants import SURVIVOR_SPAWN_RATE
from .level_constants import DEFAULT_GRID_COLS, DEFAULT_GRID_ROWS
from .models import FuelMode, Stage


def _build_stage18_reinforced_wall_zones(
    *,
    grid_cols: int,
    grid_rows: int,
    rooms_per_side: int,
    room_size: int,
    gap_width: int,
) -> list[tuple[int, int, int, int]]:
    reinforced_cells: set[tuple[int, int]] = set()

    # Inter-panel gap bands.
    inner_start = 2
    gap_cols: list[int] = []
    gap_rows: list[int] = []
    step = room_size + gap_width
    for idx in range(1, rooms_per_side):
        gap_start = inner_start + idx * step - gap_width
        gap_cols.extend(range(gap_start, gap_start + gap_width))
        gap_rows.extend(range(gap_start, gap_start + gap_width))
    for x in gap_cols:
        for y in range(inner_start, grid_rows - inner_start):
            reinforced_cells.add((x, y))
    for y in gap_rows:
        for x in range(inner_start, grid_cols - inner_start):
            reinforced_cells.add((x, y))

    # Corridor openings (width 1) through the gap bands.
    room_centers = [
        inner_start + (room_size // 2) + idx * step for idx in range(rooms_per_side)
    ]
    for y in room_centers:
        for x in gap_cols:
            reinforced_cells.discard((x, y))
    for x in room_centers:
        for y in gap_rows:
            reinforced_cells.discard((x, y))

    return [(x, y, 1, 1) for x, y in sorted(reinforced_cells)]


def _build_stage18_pitfall_zones(
    *,
    grid_cols: int,
    grid_rows: int,
    rooms_per_side: int,
    room_size: int,
    gap_width: int,
) -> list[tuple[int, int, int, int]]:
    pitfall_cells: set[tuple[int, int]] = set()
    inner_start = 2

    # Panel-edge notches: keep only the jagged concave parts as pitfalls.
    for row in range(rooms_per_side):
        for col in range(rooms_per_side):
            start_x = inner_start + col * (room_size + gap_width)
            start_y = inner_start + row * (room_size + gap_width)
            end_x = start_x + room_size - 1
            end_y = start_y + room_size - 1
            for offset in range(0, room_size, 2):
                x = start_x + offset
                y = start_y + offset
                pitfall_cells.add((x, start_y))
                pitfall_cells.add((x, end_y))
                pitfall_cells.add((start_x, y))
                pitfall_cells.add((end_x, y))

    # Keep pitfalls and reinforced walls disjoint.
    reinforced_cells = {
        (cell_x, cell_y)
        for zone_x, zone_y, zone_w, zone_h in _build_stage18_reinforced_wall_zones(
            grid_cols=grid_cols,
            grid_rows=grid_rows,
            rooms_per_side=rooms_per_side,
            room_size=room_size,
            gap_width=gap_width,
        )
        for cell_y in range(zone_y, zone_y + zone_h)
        for cell_x in range(zone_x, zone_x + zone_w)
    }
    pitfall_cells -= reinforced_cells
    pitfall_cells = {
        (x, y)
        for x, y in pitfall_cells
        if 1 <= x < grid_cols - 1 and 1 <= y < grid_rows - 1
    }
    return [(x, y, 1, 1) for x, y in sorted(pitfall_cells)]


STAGES: list[Stage] = [
    Stage(
        id="stage1",
        name_key="stages.stage1.name",
        description_key="stages.stage1.description",
        available=True,
        exterior_spawn_weight=0.97,
        interior_spawn_weight=0.03,
        zombie_normal_ratio=1.0,
    ),
    Stage(
        id="stage2",
        name_key="stages.stage2.name",
        description_key="stages.stage2.description",
        available=True,
        fuel_mode=FuelMode.FUEL_CAN,
        initial_interior_spawn_rate=0.007,
        exterior_spawn_weight=0.97,
        interior_spawn_weight=0.03,
        zombie_normal_ratio=1.0,
    ),
    Stage(
        id="stage3",
        name_key="stages.stage3.name",
        description_key="stages.stage3.description",
        available=True,
        fuel_mode=FuelMode.FUEL_CAN,
        buddy_required_count=1,
        initial_interior_spawn_rate=0.007,
        exterior_spawn_weight=0.97,
        interior_spawn_weight=0.03,
        zombie_normal_ratio=1.0,
    ),
    Stage(
        id="stage4",
        name_key="stages.stage4.name",
        description_key="stages.stage4.description",
        available=True,
        survivor_rescue_stage=True,
        waiting_car_target_count=2,
        initial_interior_spawn_rate=0.007,
        zombie_normal_ratio=1.0,
        survivor_spawn_rate=SURVIVOR_SPAWN_RATE,
    ),
    Stage(
        id="stage5",
        name_key="stages.stage5.name",
        description_key="stages.stage5.description",
        intro_key="stages.stage5.intro",
        available=True,
        wall_algorithm="default.120%",
        endurance_stage=True,
        endurance_goal_ms=1_200_000,
        fuel_spawn_count=0,
        initial_interior_spawn_rate=0.02,
        exterior_spawn_weight=0.15,
        interior_spawn_weight=0.85,
        zombie_normal_ratio=1.0,
        zombie_decay_duration_frames=int(ZOMBIE_DECAY_DURATION_FRAMES * 1.2),
    ),
    Stage(
        id="stage6",
        name_key="stages.stage6.name",
        description_key="stages.stage6.description",
        available=True,
        fuel_mode=FuelMode.FUEL_CAN,
        initial_interior_spawn_rate=0.01,
        exterior_spawn_weight=0.8,
        interior_spawn_weight=0.2,
        zombie_tracker_ratio=0.6,
        zombie_normal_ratio=0.4,
        zombie_decay_duration_frames=ZOMBIE_DECAY_DURATION_FRAMES * 2,
    ),
    Stage(
        id="stage7",
        name_key="stages.stage7.name",
        description_key="stages.stage7.description",
        available=True,
        wall_algorithm="grid_wire",
        fuel_mode=FuelMode.FUEL_CAN,
        buddy_required_count=1,
        initial_interior_spawn_rate=0.01,
        exterior_spawn_weight=0.7,
        interior_spawn_weight=0.3,
        zombie_tracker_ratio=0.3,
        zombie_wall_hugging_ratio=0.3,
        zombie_normal_ratio=0.4,
        zombie_decay_duration_frames=ZOMBIE_DECAY_DURATION_FRAMES * 2,
    ),
    Stage(
        id="stage8",
        name_key="stages.stage8.name",
        description_key="stages.stage8.description",
        available=True,
        cell_size=35,
        wall_algorithm="grid_wire",
        fuel_mode=FuelMode.FUEL_CAN,
        initial_interior_spawn_rate=0.01,
        exterior_spawn_weight=0.4,
        interior_spawn_weight=0.6,
        zombie_tracker_ratio=0.3,
        zombie_wall_hugging_ratio=0.7,
        zombie_normal_ratio=0,
        zombie_decay_duration_frames=ZOMBIE_DECAY_DURATION_FRAMES * 2,
    ),
    Stage(
        id="stage9",
        name_key="stages.stage9.name",
        description_key="stages.stage9.description",
        available=True,
        cell_size=35,
        fuel_mode=FuelMode.FUEL_CAN,
        survivor_rescue_stage=True,
        waiting_car_target_count=1,
        initial_interior_spawn_rate=0.01,
        exterior_spawn_weight=0.4,
        interior_spawn_weight=0.6,
        zombie_tracker_ratio=0.3,
        zombie_wall_hugging_ratio=0.7,
        zombie_normal_ratio=0,
        zombie_decay_duration_frames=ZOMBIE_DECAY_DURATION_FRAMES * 2,
        survivor_spawn_rate=SURVIVOR_SPAWN_RATE,
    ),
    Stage(
        id="stage10",
        name_key="stages.stage10.name",
        description_key="stages.stage10.description",
        intro_key="stages.stage10.intro",
        available=True,
        cell_size=40,
        wall_algorithm="sparse_moore.10%",
        survivor_rescue_stage=True,
        waiting_car_target_count=1,
        initial_interior_spawn_rate=0.02,
        exterior_spawn_weight=0.7,
        interior_spawn_weight=0.3,
        zombie_tracker_ratio=0.4,
        zombie_wall_hugging_ratio=0.2,
        zombie_normal_ratio=0.4,
        zombie_decay_duration_frames=ZOMBIE_DECAY_DURATION_FRAMES * 2,
        survivor_spawn_rate=0.35,
    ),
    Stage(
        id="stage11",
        name_key="stages.stage11.name",
        description_key="stages.stage11.description",
        available=True,
        grid_cols=120,
        grid_rows=7,
        wall_algorithm="sparse_moore.10%",
        initial_shoes_count=1,
        waiting_car_target_count=1,
        initial_interior_spawn_rate=0.1,
        exterior_spawn_weight=0.3,
        interior_spawn_weight=0.7,
        zombie_tracker_ratio=0.5,
        zombie_normal_ratio=0.5,
        zombie_decay_duration_frames=ZOMBIE_DECAY_DURATION_FRAMES * 2,
    ),
    Stage(
        id="stage12",
        name_key="stages.stage12.name",
        description_key="stages.stage12.description",
        available=True,
        grid_cols=32,
        grid_rows=32,
        fall_spawn_zones=[
            (4, 4, 10, 10),
            (4, 18, 10, 10),
            (18, 4, 10, 10),
            (18, 18, 10, 10),
        ],
        fuel_mode=FuelMode.FUEL_CAN,
        initial_flashlight_count=5,
        initial_shoes_count=1,
        exterior_spawn_weight=0.5,
        interior_spawn_weight=0.2,
        interior_fall_spawn_weight=0.3,
        zombie_normal_ratio=1.0,
        zombie_decay_duration_frames=ZOMBIE_DECAY_DURATION_FRAMES * 2,
    ),
    Stage(
        id="stage13",
        name_key="stages.stage13.name",
        description_key="stages.stage13.description",
        available=True,
        grid_cols=46,
        grid_rows=30,
        wall_algorithm="grid_wire",
        fall_spawn_zones=[
            (x, y, 2, 2)
            for y in range(2, DEFAULT_GRID_ROWS - 2, 4)
            for x in range(2, DEFAULT_GRID_COLS - 2, 4)
        ],
        fuel_mode=FuelMode.FUEL_CAN,
        buddy_required_count=1,
        initial_flashlight_count=3,
        initial_shoes_count=1,
        exterior_spawn_weight=0.6,
        interior_spawn_weight=0.1,
        interior_fall_spawn_weight=0.3,
        zombie_tracker_ratio=0.3,
        zombie_wall_hugging_ratio=0.3,
        zombie_normal_ratio=0.4,
        zombie_decay_duration_frames=ZOMBIE_DECAY_DURATION_FRAMES * 2,
    ),
    Stage(
        id="stage14",
        name_key="stages.stage14.name",
        description_key="stages.stage14.description",
        available=True,
        grid_cols=34,
        grid_rows=20,
        wall_rubble_ratio=0.35,
        fall_spawn_cell_ratio=0.05,
        fuel_mode=FuelMode.FUEL_CAN,
        initial_flashlight_count=3,
        initial_shoes_count=1,
        exterior_spawn_weight=0.2,
        interior_spawn_weight=0.1,
        interior_fall_spawn_weight=0.7,
        zombie_normal_ratio=1.0,
        zombie_decay_duration_frames=ZOMBIE_DECAY_DURATION_FRAMES * 2,
    ),
    Stage(
        id="stage15",
        name_key="stages.stage15.name",
        description_key="stages.stage15.description",
        intro_key="stages.stage15.intro",
        available=True,
        cell_size=35,
        grid_cols=64,
        grid_rows=24,
        wall_algorithm="grid_wire",
        fall_spawn_zones=[
            (33, 2, 4, 18),
        ],
        fuel_mode=FuelMode.FUEL_CAN,
        buddy_required_count=1,
        initial_flashlight_count=3,
        initial_shoes_count=1,
        initial_interior_spawn_rate=0.02,
        exterior_spawn_weight=0.2,
        interior_spawn_weight=0.1,
        interior_fall_spawn_weight=0.7,
        zombie_wall_hugging_ratio=0.5,
        zombie_normal_ratio=0.5,
        zombie_decay_duration_frames=ZOMBIE_DECAY_DURATION_FRAMES * 2,
    ),
    Stage(
        id="stage16",
        name_key="stages.stage16.name",
        description_key="stages.stage16.description",
        available=True,
        cell_size=60,
        grid_cols=40,
        grid_rows=25,
        wall_algorithm="sparse_moore.25%",
        pitfall_density=0.04,
        fuel_mode=FuelMode.FUEL_CAN,
        initial_flashlight_count=1,
        initial_shoes_count=1,
        initial_interior_spawn_rate=0.05,
        exterior_spawn_weight=0.7,
        interior_spawn_weight=0.3,
        zombie_normal_ratio=1.0,
        zombie_decay_duration_frames=ZOMBIE_DECAY_DURATION_FRAMES * 2,
    ),
    Stage(
        id="stage17",
        name_key="stages.stage17.name",
        description_key="stages.stage17.description",
        available=True,
        grid_cols=40,
        grid_rows=26,
        wall_algorithm="sparse_moore.25%",
        wall_rubble_ratio=0.25,
        pitfall_density=0.08,
        fuel_mode=FuelMode.FUEL_CAN,
        initial_flashlight_count=1,
        initial_shoes_count=1,
        initial_interior_spawn_rate=0.1,
        exterior_spawn_weight=0.5,
        interior_spawn_weight=0.5,
        zombie_tracker_ratio=1.0,
        zombie_decay_duration_frames=ZOMBIE_DECAY_DURATION_FRAMES * 2,
    ),
    Stage(
        id="stage18",
        name_key="stages.stage18.name",
        description_key="stages.stage18.description",
        available=True,
        grid_cols=36,
        grid_rows=36,
        wall_algorithm="sparse_ortho.30%",
        wall_rubble_ratio=0.15,
        fall_spawn_cell_ratio=0.03,
        reinforced_wall_zones=_build_stage18_reinforced_wall_zones(
            grid_cols=36,
            grid_rows=36,
            rooms_per_side=3,
            room_size=10,
            gap_width=1,
        ),
        pitfall_zones=_build_stage18_pitfall_zones(
            grid_cols=36,
            grid_rows=36,
            rooms_per_side=3,
            room_size=10,
            gap_width=1,
        ),
        fuel_mode=FuelMode.FUEL_CAN,
        initial_interior_spawn_rate=0.08,
        exterior_spawn_weight=0.6,
        interior_spawn_weight=0.4,
        zombie_tracker_ratio=0.5,
        zombie_normal_ratio=0.5,
        zombie_decay_duration_frames=ZOMBIE_DECAY_DURATION_FRAMES * 2,
    ),
    Stage(
        id="stage19",
        name_key="stages.stage19.name",
        description_key="stages.stage19.description",
        available=True,
        grid_cols=35,
        grid_rows=35,
        cell_size=35,
        wall_algorithm="grid_wire.170%",
        fall_spawn_cell_ratio=0.02,
        pitfall_density=0.008,
        fuel_mode=FuelMode.FUEL_CAN,
        buddy_required_count=1,
        initial_interior_spawn_rate=0.08,
        exterior_spawn_weight=0.4,
        interior_spawn_weight=0.0,
        interior_fall_spawn_weight=0.6,
        zombie_tracker_ratio=0.5,
        zombie_wall_hugging_ratio=0.5,
        zombie_normal_ratio=0,
        zombie_decay_duration_frames=ZOMBIE_DECAY_DURATION_FRAMES * 2,
    ),
    Stage(
        id="stage20",
        name_key="stages.stage20.name",
        description_key="stages.stage20.description",
        intro_key="stages.stage20.intro",
        available=True,
        grid_cols=40,
        grid_rows=25,
        cell_size=60,
        wall_algorithm="default.80%",
        wall_rubble_ratio=0.2,
        fall_spawn_cell_ratio=0.008,
        fall_spawn_zones=[
            (19, 10, 10, 10),
        ],
        pitfall_density=0.01,
        fuel_mode=FuelMode.FUEL_CAN,
        buddy_required_count=3,
        initial_interior_spawn_rate=0.08,
        exterior_spawn_weight=0.4,
        interior_spawn_weight=0.2,
        interior_fall_spawn_weight=0.4,
        zombie_tracker_ratio=0.4,
        zombie_wall_hugging_ratio=0.4,
        zombie_normal_ratio=0.2,
        zombie_decay_duration_frames=ZOMBIE_DECAY_DURATION_FRAMES * 2,
    ),
    Stage(
        id="stage21",
        name_key="stages.stage21.name",
        description_key="stages.stage21.description",
        available=True,
        cell_size=60,
        grid_cols=25,
        grid_rows=25,
        wall_algorithm="sparse_ortho.10%",
        pitfall_density=0.01,
        fuel_mode=FuelMode.FUEL_CAN,
        initial_interior_spawn_rate=0.05,
        exterior_spawn_weight=0.7,
        interior_spawn_weight=0.3,
        zombie_dog_ratio=0.5,
        zombie_normal_ratio=0.5,
        zombie_tracker_ratio=0.0,
        zombie_wall_hugging_ratio=0.0,
        zombie_decay_duration_frames=ZOMBIE_DECAY_DURATION_FRAMES * 2,
    ),
    Stage(
        id="stage22",
        name_key="stages.stage22.name",
        description_key="stages.stage22.description",
        available=True,
        cell_size=50,
        grid_cols=25,
        grid_rows=25,
        wall_algorithm="grid_wire.120%",
        fuel_mode=FuelMode.FUEL_CAN,
        buddy_required_count=1,
        initial_interior_spawn_rate=0.1,
        patrol_bot_spawn_rate=0.05,
        zombie_normal_ratio=1.0,
        zombie_tracker_ratio=0.0,
        zombie_wall_hugging_ratio=0.0,
        zombie_decay_duration_frames=ZOMBIE_DECAY_DURATION_FRAMES * 2,
    ),
    Stage(
        id="stage23",
        name_key="stages.stage23.name",
        description_key="stages.stage23.description",
        available=True,
        wall_algorithm="grid_wire.25%",
        pitfall_density=0.01,
        fuel_mode=FuelMode.FUEL_CAN,
        initial_interior_spawn_rate=0.06,
        patrol_bot_spawn_rate=0.04,
        exterior_spawn_weight=0.7,
        interior_spawn_weight=0.3,
        zombie_dog_ratio=1.0,
        zombie_normal_ratio=0.0,
        zombie_tracker_ratio=0.0,
        zombie_wall_hugging_ratio=0.0,
        zombie_decay_duration_frames=ZOMBIE_DECAY_DURATION_FRAMES * 2,
    ),
    Stage(
        id="stage24",
        name_key="stages.stage24.name",
        description_key="stages.stage24.description",
        available=True,
        cell_size=50,
        grid_cols=35,
        grid_rows=23,
        wall_algorithm="sparse_ortho.50%",
        fuel_mode=FuelMode.FUEL_CAN,
        initial_flashlight_count=3,
        initial_interior_spawn_rate=0.06,
        exterior_spawn_weight=0.2,
        interior_spawn_weight=0.8,
        zombie_normal_ratio=1.0,
        zombie_tracker_ratio=0.0,
        zombie_wall_hugging_ratio=0.0,
        zombie_dog_ratio=0.0,
        pitfall_zones=[(11, 9, 14, 5)],
        moving_floor_zones={
            "right": [(6, 5, 23, 1)],
            "down": [(29, 5, 1, 12)],
            "left": [(7, 17, 23, 1)],
            "up": [(6, 6, 1, 12)],
        },
        zombie_decay_duration_frames=int(ZOMBIE_DECAY_DURATION_FRAMES * 1.5),
    ),
    Stage(
        id="stage25",
        name_key="stages.stage25.name",
        description_key="stages.stage25.description",
        intro_key="stages.stage25.intro",
        available=True,
        cell_size=60,
        grid_cols=57,
        grid_rows=38,
        wall_rubble_ratio=0.5,
        wall_algorithm="normal.80%",
        fuel_mode=FuelMode.FUEL_CAN,
        initial_flashlight_count=10,
        fuel_spawn_count=1,
        patrol_bot_spawn_rate=0.01,
        initial_interior_spawn_rate=0.05,
        exterior_spawn_weight=0.05,
        interior_spawn_weight=0.95,
        zombie_decay_duration_frames=ZOMBIE_DECAY_DURATION_FRAMES * 3,
        zombie_normal_ratio=0.4,
        zombie_tracker_ratio=0.2,
        zombie_wall_hugging_ratio=0,
        zombie_dog_ratio=0.4,
        moving_floor_zones={
            "up": [
                (8, 3, 1, 14),
                (21, 3, 1, 14),
                (34, 3, 1, 14),
                (47, 3, 1, 14),
                (8, 21, 1, 14),
                (21, 21, 1, 14),
                (34, 21, 1, 14),
                (47, 21, 1, 14),
            ],
            "down": [
                (9, 3, 1, 14),
                (22, 3, 1, 14),
                (35, 3, 1, 14),
                (48, 3, 1, 14),
                (9, 21, 1, 14),
                (22, 21, 1, 14),
                (35, 21, 1, 14),
                (48, 21, 1, 14),
            ],
            "left": [
                (3, 19, 5, 1),
                (10, 19, 5, 1),
                (16, 19, 5, 1),
                (23, 19, 5, 1),
                (29, 19, 5, 1),
                (36, 19, 5, 1),
                (42, 19, 5, 1),
                (49, 19, 5, 1),
            ],
            "right": [
                (3, 18, 5, 1),
                (10, 18, 5, 1),
                (16, 18, 5, 1),
                (23, 18, 5, 1),
                (29, 18, 5, 1),
                (36, 18, 5, 1),
                (42, 18, 5, 1),
                (49, 18, 5, 1),
            ],
        },
        fall_spawn_zones=[
            (2, 2, 53, 1),
            (2, 3, 1, 33),
            (15, 3, 1, 33),
            (28, 3, 1, 33),
            (41, 3, 1, 33),
            (54, 3, 1, 33),
            (3, 35, 12, 1),
            (16, 35, 12, 1),
            (29, 35, 12, 1),
            (42, 35, 12, 1),
        ],
        buddy_required_count=1,
        survivor_spawn_rate=0.07,
        exit_sides=["top", "bottom"],
    ),
    Stage(
        id="stage26",
        name_key="stages.stage26.name",
        description_key="stages.stage26.description",
        available=True,
        cell_size=50,
        grid_cols=30,
        grid_rows=25,
        wall_algorithm="sparse_ortho.10%",
        fuel_mode=FuelMode.REFUEL_CHAIN,
        buddy_required_count=1,
        initial_interior_spawn_rate=0.05,
        exterior_spawn_weight=0.2,
        interior_spawn_weight=0.0,
        interior_fall_spawn_weight=0.8,
        zombie_normal_ratio=1.0,
        zombie_tracker_ratio=0.0,
        zombie_wall_hugging_ratio=0.0,
        zombie_dog_ratio=0.0,
        initial_flashlight_count=2,
        initial_shoes_count=2,
        waiting_car_target_count=1,
        zombie_decay_duration_frames=ZOMBIE_DECAY_DURATION_FRAMES * 3,
        moving_floor_zones={
            "up": [],
            "down": [
                (5, 7, 5, 1),
                (11, 7, 5, 1),
                (17, 7, 5, 1),
                (23, 7, 5, 1),
                (5, 12, 5, 1),
                (11, 12, 5, 1),
                (17, 12, 5, 1),
                (23, 12, 5, 1),
                (5, 16, 2, 1),
                (8, 16, 2, 1),
                (11, 16, 2, 1),
                (14, 16, 2, 1),
                (17, 16, 2, 1),
                (20, 16, 2, 1),
                (23, 16, 2, 1),
                (26, 16, 2, 1),
                (5, 20, 2, 1),
                (8, 20, 2, 1),
                (11, 20, 2, 1),
                (14, 20, 2, 1),
                (17, 20, 2, 1),
                (20, 20, 2, 1),
                (23, 20, 2, 1),
                (26, 20, 2, 1),
            ],
            "left": [],
            "right": [],
        },
        pitfall_zones=[
            (28, 2, 1, 19),
            (4, 3, 1, 18),
            (10, 7, 1, 1),
            (16, 7, 1, 14),
            (22, 7, 1, 1),
            (10, 12, 1, 9),
            (22, 12, 1, 9),
            (7, 16, 1, 5),
            (13, 16, 1, 5),
            (19, 16, 1, 5),
            (25, 16, 1, 5),
        ],
        fall_spawn_zones=[
            (5, 4, 2, 2),
            (8, 4, 2, 2),
            (11, 4, 2, 2),
            (14, 4, 2, 2),
            (17, 4, 2, 2),
            (20, 4, 2, 2),
            (23, 4, 2, 2),
            (26, 4, 2, 2),
        ],
    ),
    Stage(
        id="stage27",
        name_key="stages.stage27.name",
        description_key="stages.stage27.description",
        available=True,
        cell_size=60,
        grid_cols=30,
        grid_rows=30,
        wall_algorithm="sparse_ortho.20%",
        fuel_mode=FuelMode.FUEL_CAN,
        initial_interior_spawn_rate=0.1,
        exterior_spawn_weight=0.2,
        interior_spawn_weight=0.8,
        zombie_normal_ratio=0.125,
        zombie_lineformer_ratio=0.75,
        zombie_tracker_ratio=0.125,
        zombie_wall_hugging_ratio=0.0,
        zombie_dog_ratio=0.0,
        initial_flashlight_count=2,
        initial_shoes_count=0,
        waiting_car_target_count=1,
        zombie_decay_duration_frames=ZOMBIE_DECAY_DURATION_FRAMES * 3,
        zombie_spawn_count_per_interval=4,
        survivor_spawn_rate=0.02,
    ),
    Stage(
        id="stage28",
        name_key="stages.stage28.name",
        description_key="stages.stage28.description",
        fuel_mode=FuelMode.REFUEL_CHAIN,
        available=True,
        cell_size=60,
        grid_cols=39,
        grid_rows=14,
        wall_algorithm="empty",
        initial_interior_spawn_rate=0.05,
        exterior_spawn_weight=0.2,
        interior_spawn_weight=0.2,
        interior_fall_spawn_weight=0.6,
        zombie_normal_ratio=0.5,
        zombie_dog_ratio=0.5,
        initial_flashlight_count=0,
        initial_shoes_count=2,
        waiting_car_target_count=1,
        zombie_decay_duration_frames=ZOMBIE_DECAY_DURATION_FRAMES * 3,
        moving_floor_zones={
            "up": [],
            "down": [],
            "left": [(3, 10, 33, 1)],
            "right": [(3, 3, 33, 1)],
        },
        fall_spawn_zones=[
            (3, 2, 3, 1),
            (8, 2, 3, 1),
            (13, 2, 3, 1),
            (18, 2, 3, 1),
            (23, 2, 3, 1),
            (28, 2, 3, 1),
            (33, 2, 3, 1),
            (3, 11, 3, 1),
            (8, 11, 3, 1),
            (13, 11, 3, 1),
            (18, 11, 3, 1),
            (23, 11, 3, 1),
            (28, 11, 3, 1),
            (33, 11, 3, 1),
        ],
        puddle_zones=[
            (2, 3, 1, 8),
            (36, 3, 1, 8),
            (6, 4, 2, 6),
            (11, 4, 2, 6),
            (16, 4, 2, 6),
            (21, 4, 2, 6),
            (26, 4, 2, 6),
            (31, 4, 2, 6),
        ],
        puddle_density=0.05,
        exit_sides=["top", "bottom"],
        spiky_plant_density=0.04,
    ),
    Stage(
        id="stage29",
        name_key="stages.stage29.name",
        description_key="stages.stage29.description",
        available=True,
        fuel_mode=FuelMode.FUEL_CAN,
        cell_size=40,
        grid_cols=44,
        grid_rows=44,
        wall_algorithm="empty",
        buddy_required_count=1,
        initial_interior_spawn_rate=0.1,
        exterior_spawn_weight=0.2,
        interior_spawn_weight=0.8,
        zombie_normal_ratio=1.0,
        zombie_dog_ratio=0.0,
        zombie_lineformer_ratio=0.0,
        initial_flashlight_count=3,
        waiting_car_target_count=1,
        zombie_decay_duration_frames=ZOMBIE_DECAY_DURATION_FRAMES * 3,
        spiky_plant_zones=[
            (3, 3, 38, 2),
            (3, 5, 2, 36),
            (7, 5, 2, 36),
            (11, 5, 2, 36),
            (15, 5, 2, 36),
            (19, 5, 2, 36),
            (23, 5, 2, 36),
            (27, 5, 2, 36),
            (31, 5, 2, 36),
            (35, 5, 2, 36),
            (39, 5, 2, 36),
            (5, 7, 2, 2),
            (9, 7, 2, 2),
            (13, 7, 2, 2),
            (17, 7, 2, 2),
            (21, 7, 2, 2),
            (25, 7, 2, 2),
            (29, 7, 2, 2),
            (33, 7, 2, 2),
            (37, 7, 2, 2),
            (5, 11, 2, 2),
            (9, 11, 2, 2),
            (13, 11, 2, 2),
            (17, 11, 2, 2),
            (21, 11, 2, 2),
            (25, 11, 2, 2),
            (29, 11, 2, 2),
            (33, 11, 2, 2),
            (37, 11, 2, 2),
            (5, 15, 2, 2),
            (9, 15, 2, 2),
            (13, 15, 2, 2),
            (17, 15, 2, 2),
            (21, 15, 2, 2),
            (25, 15, 2, 2),
            (29, 15, 2, 2),
            (33, 15, 2, 2),
            (37, 15, 2, 2),
            (5, 19, 2, 2),
            (9, 19, 2, 2),
            (13, 19, 2, 2),
            (17, 19, 2, 2),
            (21, 19, 2, 2),
            (25, 19, 2, 2),
            (29, 19, 2, 2),
            (33, 19, 2, 2),
            (37, 19, 2, 2),
            (5, 23, 2, 2),
            (9, 23, 2, 2),
            (13, 23, 2, 2),
            (17, 23, 2, 2),
            (21, 23, 2, 2),
            (25, 23, 2, 2),
            (29, 23, 2, 2),
            (33, 23, 2, 2),
            (37, 23, 2, 2),
            (5, 27, 2, 2),
            (9, 27, 2, 2),
            (13, 27, 2, 2),
            (17, 27, 2, 2),
            (21, 27, 2, 2),
            (25, 27, 2, 2),
            (29, 27, 2, 2),
            (33, 27, 2, 2),
            (37, 27, 2, 2),
            (5, 31, 2, 2),
            (9, 31, 2, 2),
            (13, 31, 2, 2),
            (17, 31, 2, 2),
            (21, 31, 2, 2),
            (25, 31, 2, 2),
            (29, 31, 2, 2),
            (33, 31, 2, 2),
            (37, 31, 2, 2),
            (5, 35, 2, 2),
            (9, 35, 2, 2),
            (13, 35, 2, 2),
            (17, 35, 2, 2),
            (21, 35, 2, 2),
            (25, 35, 2, 2),
            (29, 35, 2, 2),
            (33, 35, 2, 2),
            (37, 35, 2, 2),
            (5, 39, 2, 2),
            (9, 39, 2, 2),
            (13, 39, 2, 2),
            (17, 39, 2, 2),
            (21, 39, 2, 2),
            (25, 39, 2, 2),
            (29, 39, 2, 2),
            (33, 39, 2, 2),
            (37, 39, 2, 2),
        ],
        puddle_density=0.05,
    ),
    Stage(
        id="stage30",
        name_key="stages.stage30.name",
        description_key="stages.stage30.description",
        intro_key="stages.stage30.intro",
        available=True,
        cell_size=40,
        grid_cols=40,
        grid_rows=25,
        wall_algorithm="corridor.15%",
        initial_interior_spawn_rate=0.08,
        exterior_spawn_weight=0.4,
        interior_spawn_weight=0.6,
        zombie_normal_ratio=0.2,
        zombie_lineformer_ratio=0.4,
        zombie_tracker_ratio=0.2,
        zombie_wall_hugging_ratio=0.2,
        zombie_dog_ratio=0.0,
        initial_flashlight_count=3,
        initial_shoes_count=2,
        waiting_car_target_count=0,
        zombie_decay_duration_frames=ZOMBIE_DECAY_DURATION_FRAMES * 2,
        endurance_stage=True,
        endurance_goal_ms=900_000,
        fuel_spawn_count=0,
        survivor_spawn_rate=0.05,
    ),
    Stage(
        id="stage31",
        name_key="stages.stage31.name",
        description_key="stages.stage31.description",
        fuel_mode=FuelMode.REFUEL_CHAIN,
        available=True,
        cell_size=50,
        grid_cols=35,
        grid_rows=35,
        wall_algorithm="empty",
        reinforced_wall_density=0.25,
        initial_interior_spawn_rate=0.02,
        exterior_spawn_weight=0.4,
        interior_spawn_weight=0.6,
        zombie_normal_ratio=0.0,
        zombie_dog_ratio=0.3,
        zombie_tracker_ratio=0.0,
        zombie_wall_hugging_ratio=0.0,
        zombie_lineformer_ratio=0.0,
        zombie_nimble_dog_ratio=0.7,
        zombie_decay_duration_frames=ZOMBIE_DECAY_DURATION_FRAMES * 2,
        patrol_bot_spawn_rate=0.02,
    ),
    Stage(
        id="stage32",
        name_key="stages.stage32.name",
        description_key="stages.stage32.description",
        fuel_mode=FuelMode.REFUEL_CHAIN,
        available=True,
        cell_size=50,
        grid_cols=20,
        grid_rows=53,
        wall_algorithm="empty",
        reinforced_wall_density=0.15,
        initial_interior_spawn_rate=0.03,
        exterior_spawn_weight=0.4,
        interior_spawn_weight=0.6,
        zombie_normal_ratio=0.5,
        zombie_dog_ratio=0.0,
        zombie_tracker_ratio=0.4,
        zombie_wall_hugging_ratio=0.0,
        zombie_lineformer_ratio=0.0,
        zombie_nimble_dog_ratio=0.1,
        zombie_decay_duration_frames=ZOMBIE_DECAY_DURATION_FRAMES * 2,
        fire_floor_zones=[
            (3, 11, 14, 1),
            (3, 21, 14, 1),
            (3, 31, 14, 1),
            (3, 41, 14, 1),
        ],
        exit_sides=["top", "bottom"],
    ),
    Stage(
        id="stage33",
        name_key="stages.stage33.name",
        description_key="stages.stage33.description",
        fuel_mode=FuelMode.REFUEL_CHAIN,
        available=True,
        cell_size=50,
        grid_cols=30,
        grid_rows=24,
        wall_algorithm="empty",
        initial_interior_spawn_rate=0.08,
        exterior_spawn_weight=0.2,
        interior_spawn_weight=0.5,
        interior_fall_spawn_weight=0.3,
        fall_spawn_cell_ratio=0.08,
        zombie_normal_ratio=0.2,
        zombie_lineformer_ratio=0.0,
        zombie_tracker_ratio=0.0,
        zombie_wall_hugging_ratio=0.0,
        zombie_dog_ratio=0.0,
        zombie_nimble_dog_ratio=0.8,
        initial_flashlight_count=5,
        initial_shoes_count=2,
        moving_floor_zones={
            "down": [
                (3, 3, 4, 1),
                (8, 3, 4, 1),
                (13, 3, 4, 1),
                (18, 3, 4, 1),
                (23, 3, 4, 1),
                (3, 5, 4, 1),
                (8, 5, 4, 1),
                (13, 5, 4, 1),
                (18, 5, 4, 1),
                (23, 5, 4, 1),
                (3, 7, 4, 1),
                (8, 7, 4, 1),
                (13, 7, 4, 1),
                (18, 7, 4, 1),
                (23, 7, 4, 1),
                (3, 9, 4, 1),
                (8, 9, 4, 1),
                (13, 9, 4, 1),
                (18, 9, 4, 1),
                (23, 9, 4, 1),
                (3, 11, 4, 1),
                (8, 11, 4, 1),
                (13, 11, 4, 1),
                (18, 11, 4, 1),
                (23, 11, 4, 1),
                (3, 13, 4, 1),
                (8, 13, 4, 1),
                (13, 13, 4, 1),
                (18, 13, 4, 1),
                (23, 13, 4, 1),
                (3, 15, 4, 1),
                (8, 15, 4, 1),
                (13, 15, 4, 1),
                (18, 15, 4, 1),
                (23, 15, 4, 1),
                (3, 17, 4, 1),
                (8, 17, 4, 1),
                (13, 17, 4, 1),
                (18, 17, 4, 1),
                (23, 17, 4, 1),
                (3, 19, 4, 1),
                (8, 19, 4, 1),
                (13, 19, 4, 1),
                (18, 19, 4, 1),
                (23, 19, 4, 1),
            ],
        },
        reinforced_wall_zones=[
            (7, 3, 1, 1),
            (12, 3, 1, 1),
            (17, 3, 1, 1),
            (22, 3, 1, 1),
            (27, 3, 1, 1),
            (2, 5, 1, 1),
            (7, 5, 1, 1),
            (12, 5, 1, 1),
            (17, 5, 1, 1),
            (22, 5, 1, 1),
            (7, 7, 1, 1),
            (12, 7, 1, 1),
            (17, 7, 1, 1),
            (22, 7, 1, 1),
            (27, 7, 1, 1),
            (2, 9, 1, 1),
            (7, 9, 1, 1),
            (12, 9, 1, 1),
            (17, 9, 1, 1),
            (22, 9, 1, 1),
            (7, 11, 1, 1),
            (12, 11, 1, 1),
            (17, 11, 1, 1),
            (22, 11, 1, 1),
            (27, 11, 1, 1),
            (2, 13, 1, 1),
            (7, 13, 1, 1),
            (12, 13, 1, 1),
            (17, 13, 1, 1),
            (22, 13, 1, 1),
            (7, 15, 1, 1),
            (12, 15, 1, 1),
            (17, 15, 1, 1),
            (22, 15, 1, 1),
            (27, 15, 1, 1),
            (2, 17, 1, 1),
            (7, 17, 1, 1),
            (12, 17, 1, 1),
            (17, 17, 1, 1),
            (22, 17, 1, 1),
            (7, 19, 1, 1),
            (12, 19, 1, 1),
            (17, 19, 1, 1),
            (22, 19, 1, 1),
            (27, 19, 1, 1),
        ],
        pitfall_zones=[
            (2, 21, 26, 1),
        ],
        exit_sides=["top", "bottom"],
        zombie_decay_duration_frames=ZOMBIE_DECAY_DURATION_FRAMES * 3,
    ),
    Stage(
        id="stage34",
        name_key="stages.stage34.name",
        description_key="stages.stage34.description",
        available=True,
        cell_size=50,
        grid_cols=30,
        grid_rows=30,
        wall_algorithm="sparse_ortho.30%",
        fuel_mode=FuelMode.REFUEL_CHAIN,
        initial_interior_spawn_rate=0.2,
        exterior_spawn_weight=0.4,
        interior_spawn_weight=0.6,
        interior_fall_spawn_weight=0.0,
        zombie_normal_ratio=0.25,
        zombie_tracker_ratio=0.25,
        zombie_wall_hugging_ratio=0.25,
        zombie_lineformer_ratio=0.25,
        zombie_solitary_ratio=0.0,
        zombie_dog_ratio=0.0,
        zombie_nimble_dog_ratio=0.0,
        initial_flashlight_count=1,
        initial_shoes_count=1,
        reinforced_wall_density=0.18,
        fire_floor_density=0.18,
        patrol_bot_spawn_rate=0.02,
    ),
    # Stage(
    #     id="stage34",
    #     name_key="stages.stage34.name",
    #     description_key="stages.stage34.description",
    #     fuel_mode=FuelMode.REFUEL_CHAIN,
    #     available=False,
    #     cell_size=50,
    #     grid_cols=30,
    #     grid_rows=30,
    #     wall_algorithm="grid_wire.40%",
    #     fall_spawn_cell_ratio=0.04,
    #     reinforced_wall_density=0.12,
    #     initial_interior_spawn_rate=0.1,
    #     exterior_spawn_weight=0.1,
    #     interior_spawn_weight=0.6,
    #     interior_fall_spawn_weight=0.3,
    #     zombie_normal_ratio=0.6,
    #     zombie_tracker_ratio=0.0,
    #     zombie_wall_hugging_ratio=0.0,
    #     zombie_lineformer_ratio=0.0,
    #     zombie_solitary_ratio=0.4,
    #     zombie_dog_ratio=0.0,
    #     zombie_nimble_dog_ratio=0.0,
    #     zombie_spawn_count_per_interval=2,
    #     zombie_decay_duration_frames=ZOMBIE_DECAY_DURATION_FRAMES * 3,
    #     pitfall_density=0.01,
    # ),
    Stage(
        id="stage35",
        name_key="stages.stage35.name",
        description_key="stages.stage35.description",
        intro_key="stages.stage35.intro",
        available=True,
        cell_size=40,
        grid_cols=25,
        grid_rows=25,
        wall_algorithm="sparse_moore.20%",
        wall_rubble_ratio=0.5,
        reinforced_wall_density=0.05,
        initial_interior_spawn_rate=0.12,
        exterior_spawn_weight=0.7,
        interior_spawn_weight=0.27,
        interior_fall_spawn_weight=0.03,
        zombie_normal_ratio=0.6,
        zombie_lineformer_ratio=0.2,
        zombie_tracker_ratio=0.2,
        zombie_wall_hugging_ratio=0.0,
        zombie_dog_ratio=0.0,
        initial_flashlight_count=3,
        initial_shoes_count=2,
        waiting_car_target_count=1,
        zombie_decay_duration_frames=ZOMBIE_DECAY_DURATION_FRAMES * 3,
        spiky_plant_density=0.12,
        puddle_density=0.01,
        # pitfall_zones=[(12, 12, 1, 1)],
        spiky_plant_zones=[
            (12, 11, 1, 1),
            (11, 12, 1, 1),
            (13, 12, 1, 1),
            (12, 13, 1, 1),
        ],
        fall_spawn_zones=[
            (3, 3, 1, 1),
            (9, 3, 1, 1),
            (15, 3, 1, 1),
            (21, 3, 1, 1),
            (3, 21, 1, 1),
            (9, 21, 1, 1),
            (15, 21, 1, 1),
            (21, 21, 1, 1),
            (3, 3, 1, 1),
            (3, 9, 1, 1),
            (3, 15, 1, 1),
            (3, 21, 1, 1),
            (21, 3, 1, 1),
            (21, 9, 1, 1),
            (21, 15, 1, 1),
            (21, 21, 1, 1),
        ],
        zombie_spawn_count_per_interval=3,
        endurance_stage=True,
        endurance_goal_ms=600_000,
    ),
]
DEFAULT_STAGE_ID = "stage1"


__all__ = [
    "STAGES",
    "DEFAULT_STAGE_ID",
]



from __future__ import annotations

import math
import random
from typing import Any, Iterable, Protocol

import pygame

try:
    from typing import Self
except ImportError:  # pragma: no cover - Python 3.10 fallback
    from typing_extensions import Self

from ..entities_constants import (
    FAST_ZOMBIE_BASE_SPEED,
    ZombieKind,
    ZOMBIE_CARBONIZE_DECAY_FRAMES,
    ZOMBIE_DECAY_DURATION_FRAMES,
    ZOMBIE_DECAY_MIN_SPEED_RATIO,
    ZOMBIE_DOG_LONG_AXIS_RATIO,
    ZOMBIE_DOG_SHORT_AXIS_RATIO,
    PUDDLE_SPEED_FACTOR,
    ZOMBIE_CONTAMINATED_SPEED_FACTOR,
    ZOMBIE_RADIUS,
    ZOMBIE_SEPARATION_DISTANCE,
    ZOMBIE_SPEED,
    ZOMBIE_TRACKER_SCAN_INTERVAL_MS,
    ZOMBIE_TRACKER_WANDER_INTERVAL_MS,
    ZOMBIE_WALL_DAMAGE,
    ZOMBIE_WANDER_INTERVAL_MS,
    PATROL_BOT_ZOMBIE_DAMAGE,
    PATROL_BOT_ZOMBIE_DAMAGE_INTERVAL_FRAMES,
    PATROL_BOT_PARALYZE_MS,
    PATROL_BOT_PARALYZE_BLINK_MS,
    PATROL_BOT_PARALYZE_MARKER_COLOR,
    TRAPPED_ZOMBIE_SLOW_FACTOR,
    TRAPPED_ZOMBIE_REPEL_MAX_MULT,
    TRAPPED_ZOMBIE_REPEL_PER_STACK,
    TRAPPED_ZOMBIE_REPEL_RADIUS_CELLS,
)
from ..models import Footprint, LevelLayout
from ..render.entity_overlays import (
    apply_zombie_kind_overlay,
    draw_paralyze_marker_overlay,
)
from ..render_constants import ENTITY_SHADOW_RADIUS_MULT
from ..render_assets import (
    angle_bin_from_vector,
    build_zombie_directional_surfaces,
    build_zombie_dog_directional_surfaces,
)
from ..rng import get_rng
from ..surface_effects import SpikyPlantLike, is_in_contaminated_cell, is_in_puddle_cell
from ..screen_constants import SCREEN_HEIGHT, SCREEN_WIDTH
from ..world_grid import apply_cell_edge_nudge
from .movement import _circle_wall_collision
from .zombie_movement import (
    _zombie_lineformer_train_head_movement,
    _zombie_solitary_movement,
    _zombie_normal_movement,
    _zombie_tracker_movement,
    _zombie_wall_hug_movement,
)
from .walls import Wall
from .zombie_visuals import build_grayscale_image
from .zombie_vitals import ZombieVitals

RNG = get_rng()


class MovementStrategy(Protocol):
    def __call__(
        self,
        zombie: "Zombie",
        cell_size: int,
        layout: LevelLayout,
        player_center: tuple[float, float],
        nearby_zombies: Iterable["Zombie"],
        footprints: list[Footprint],
        *,
        now_ms: int,
    ) -> tuple[float, float]: ...


class Zombie(pygame.sprite.Sprite):
    _next_lineformer_id = 1

    def __init__(
        self: Self,
        x: float,
        y: float,
        *,
        speed: float = ZOMBIE_SPEED,
        kind: ZombieKind = ZombieKind.NORMAL,
        movement_strategy: MovementStrategy | None = None,
        decay_duration_frames: float = ZOMBIE_DECAY_DURATION_FRAMES,
    ) -> None:
        super().__init__()
        self.radius = ZOMBIE_RADIUS
        self.facing_bin = 0
        self.kind = kind
        self.directional_images = build_zombie_directional_surfaces(
            self.radius,
            draw_hands=False,
        )
        self.image = self.directional_images[self.facing_bin]
        self.rect = self.image.get_rect(center=(x, y))
        jitter_base = FAST_ZOMBIE_BASE_SPEED if speed > ZOMBIE_SPEED else ZOMBIE_SPEED
        jitter = jitter_base * 0.2
        base_speed = speed + RNG.uniform(-jitter, jitter)
        self.initial_speed = base_speed
        self.speed = base_speed
        self.x = float(self.rect.centerx)
        self.y = float(self.rect.centery)
        self.was_in_sight = False
        self.vitals = ZombieVitals(
            max_health=100,
            decay_duration_frames=decay_duration_frames,
            decay_min_speed_ratio=ZOMBIE_DECAY_MIN_SPEED_RATIO,
            carbonize_decay_frames=ZOMBIE_CARBONIZE_DECAY_FRAMES,
            on_health_ratio=self._apply_speed_ratio,
            on_kill=self.kill,
            on_carbonize=self._apply_carbonize_visuals,
        )
        if movement_strategy is None:
            if self.kind == ZombieKind.TRACKER:
                movement_strategy = _zombie_tracker_movement
            elif self.kind == ZombieKind.WALL_HUGGER:
                movement_strategy = _zombie_wall_hug_movement
            elif self.kind == ZombieKind.LINEFORMER:
                movement_strategy = _zombie_lineformer_train_head_movement
            elif self.kind == ZombieKind.SOLITARY:
                movement_strategy = _zombie_solitary_movement
            else:
                movement_strategy = _zombie_normal_movement
        self.movement_strategy = movement_strategy
        self.lineformer_id = Zombie._next_lineformer_id
        Zombie._next_lineformer_id += 1
        self.lineformer_follow_target_id: int | None = None
        self.lineformer_target_pos: tuple[float, float] | None = None
        self.lineformer_last_target_seen_ms: int | None = None
        self.tracker_target_pos: tuple[float, float] | None = None
        self.tracker_target_time: int | None = None
        self.tracker_last_scan_time = 0
        self.tracker_scan_interval_ms = ZOMBIE_TRACKER_SCAN_INTERVAL_MS
        self.tracker_relock_after_time: int | None = None
        self.tracker_force_wander = False
        if self.kind == ZombieKind.WALL_HUGGER:
            self.wall_hug_side = RNG.choice([-1.0, 1.0])
            self.wall_hug_angle = RNG.uniform(0, math.tau)
        else:
            self.wall_hug_side = 0.0
            self.wall_hug_angle = None
        self.wall_hug_last_wall_time: int | None = None
        self.wall_hug_last_side_has_wall = False
        self.wall_hug_stuck_flag = False
        self.wander_angle = RNG.uniform(0, math.tau)
        self.wander_interval_ms = (
            ZOMBIE_TRACKER_WANDER_INTERVAL_MS
            if self.kind == ZombieKind.TRACKER
            else ZOMBIE_WANDER_INTERVAL_MS
        )
        self.last_wander_change_time = 0
        self.wander_change_interval = max(
            0, self.wander_interval_ms + RNG.randint(-500, 500)
        )
        self.last_move_dx = 0.0
        self.last_move_dy = 0.0
        self.collision_radius = float(self.radius)
        self.shadow_radius = max(
            1, int(self.collision_radius * ENTITY_SHADOW_RADIUS_MULT)
        )
        self.shadow_offset_scale = 1.0
        self.solitary_eval_frame_counter = 0
        self.solitary_committed_move: tuple[int, int] | None = None
        self.solitary_previous_move: tuple[int, int] | None = None

    @property
    def max_health(self: Self) -> int:
        return self.vitals.max_health

    @property
    def health(self: Self) -> int:
        return self.vitals.health

    @property
    def decay_duration_frames(self: Self) -> float:
        return self.vitals.decay_duration_frames

    @property
    def carbonized(self: Self) -> bool:
        return self.vitals.carbonized

    @property
    def last_damage_ms(self: Self) -> int | None:
        return self.vitals.last_damage_ms

    @property
    def last_damage_source(self: Self) -> str | None:
        return self.vitals.last_damage_source

    @property
    def patrol_paralyze_until_ms(self: Self) -> int:
        return self.vitals.patrol_paralyze_until_ms

    @property
    def patrol_damage_frame_counter(self: Self) -> int:
        return self.vitals.patrol_damage_frame_counter

    def _apply_speed_ratio(self: Self, ratio: float) -> None:
        self.speed = self.initial_speed * ratio

    def _apply_carbonize_visuals(self: Self) -> None:
        self.image = build_grayscale_image(self.image)

    def _update_mode(
        self: Self, player_center: tuple[float, float], sight_range: float
    ) -> bool:
        dx_target = player_center[0] - self.x
        dy_target = player_center[1] - self.y
        dist_to_player_sq = dx_target * dx_target + dy_target * dy_target
        is_in_sight = dist_to_player_sq <= sight_range * sight_range
        self.was_in_sight = is_in_sight
        return is_in_sight

    def _handle_wall_collision(
        self: Self, next_x: float, next_y: float, walls: list[Wall]
    ) -> tuple[float, float]:
        final_x, final_y = next_x, next_y

        possible_walls = [
            w
            for w in walls
            if abs(w.rect.centerx - self.x) < 100 and abs(w.rect.centery - self.y) < 100
        ]

        for wall in possible_walls:
            collides = _circle_wall_collision(
                (next_x, self.y), self.collision_radius, wall
            )
            if collides:
                if wall.alive():
                    wall._take_damage(amount=ZOMBIE_WALL_DAMAGE)
                if wall.alive():
                    final_x = self.x
                    break

        for wall in possible_walls:
            collides = _circle_wall_collision(
                (final_x, next_y), self.collision_radius, wall
            )
            if collides:
                if wall.alive():
                    wall._take_damage(amount=ZOMBIE_WALL_DAMAGE)
                if wall.alive():
                    final_y = self.y
                    break

        return final_x, final_y

    def _avoid_other_zombies(
        self: Self,
        move_x: float,
        move_y: float,
        zombies: Iterable[pygame.sprite.Sprite],
    ) -> tuple[float, float]:
        """If another zombie is too close, steer directly away from the closest one."""
        orig_move_x, orig_move_y = move_x, move_y
        next_x = self.x + move_x
        next_y = self.y + move_y

        closest: pygame.sprite.Sprite | None = None
        closest_dist_sq = ZOMBIE_SEPARATION_DISTANCE * ZOMBIE_SEPARATION_DISTANCE
        for other in zombies:
            if other is self or not other.alive():
                continue
            if getattr(other, "is_trapped", False):
                continue

            # Lineformer logic: non-lineformers ignore lineformers
            other_kind = other.kind  # type: ignore[attr-defined]
            if (
                self.kind != ZombieKind.LINEFORMER
                and other_kind == ZombieKind.LINEFORMER
            ):
                continue

            # Type ignore because spatial index might contain other sprites,
            # but we only query ZOMBIE | ZOMBIE_DOG | TRAPPED_ZOMBIE
            ox = other.x  # type: ignore[attr-defined]
            oy = other.y  # type: ignore[attr-defined]

            dx = ox - next_x
            dy = oy - next_y
            if (
                abs(dx) > ZOMBIE_SEPARATION_DISTANCE
                or abs(dy) > ZOMBIE_SEPARATION_DISTANCE
            ):
                continue
            dist_sq = dx * dx + dy * dy
            if dist_sq < closest_dist_sq:
                closest = other
                closest_dist_sq = dist_sq

        if closest is None:
            return move_x, move_y

        if self.kind == ZombieKind.WALL_HUGGER:
            other_radius = float(closest.collision_radius)  # type: ignore[attr-defined]
            bump_dist_sq = (self.collision_radius + other_radius) ** 2
            if closest_dist_sq < bump_dist_sq and RNG.random() < 0.1:
                if self.wall_hug_angle is None:
                    self.wall_hug_angle = self.wander_angle
                self.wall_hug_angle = (self.wall_hug_angle + math.pi) % math.tau
                self.wall_hug_side *= -1.0
                return (
                    math.cos(self.wall_hug_angle) * self.speed,
                    math.sin(self.wall_hug_angle) * self.speed,
                )

        away_dx = next_x - closest.x  # type: ignore[attr-defined]
        away_dy = next_y - closest.y  # type: ignore[attr-defined]
        away_dist = math.hypot(away_dx, away_dy)
        if away_dist == 0:
            angle = RNG.uniform(0, 2 * math.pi)
            away_dx, away_dy = math.cos(angle), math.sin(angle)
            away_dist = 1

        move_x = (away_dx / away_dist) * self.speed
        move_y = (away_dy / away_dist) * self.speed
        if self.kind == ZombieKind.WALL_HUGGER:
            if orig_move_x or orig_move_y:
                orig_angle = math.atan2(orig_move_y, orig_move_x)
                new_angle = math.atan2(move_y, move_x)
                diff = (new_angle - orig_angle + math.pi) % math.tau - math.pi
                if abs(diff) > math.pi / 2.0:
                    clamped = math.copysign(math.pi / 2.0, diff)
                    new_angle = orig_angle + clamped
                    move_x = math.cos(new_angle) * self.speed
                    move_y = math.sin(new_angle) * self.speed
        return move_x, move_y

    def _slow_near_trapped_zombies(
        self: Self,
        move_x: float,
        move_y: float,
        zombies: Iterable[pygame.sprite.Sprite],
    ) -> tuple[float, float]:
        next_x = self.x + move_x
        next_y = self.y + move_y
        for other in zombies:
            if other is self or not other.alive():
                continue
            if not getattr(other, "is_trapped", False):
                continue
            ox = float(getattr(other, "x", other.rect.centerx))
            oy = float(getattr(other, "y", other.rect.centery))
            other_radius = float(getattr(other, "collision_radius", 0.0))
            dx = ox - next_x
            dy = oy - next_y
            touch_radius = self.collision_radius + other_radius
            if dx * dx + dy * dy <= touch_radius * touch_radius:
                return (
                    move_x * TRAPPED_ZOMBIE_SLOW_FACTOR,
                    move_y * TRAPPED_ZOMBIE_SLOW_FACTOR,
                )
        return move_x, move_y

    def _repel_from_loaded_spiky_plants(
        self: Self,
        move_x: float,
        move_y: float,
        *,
        cell_size: int,
        spiky_plants: dict[tuple[int, int], SpikyPlantLike] | None,
        trapped_spiky_plant_counts: dict[tuple[int, int], int] | None,
    ) -> tuple[float, float]:
        if (
            cell_size <= 0
            or not spiky_plants
            or not trapped_spiky_plant_counts
        ):
            return move_x, move_y
        next_x = self.x + move_x
        next_y = self.y + move_y
        effect_radius = max(
            self.collision_radius * 2.0,
            float(cell_size) * TRAPPED_ZOMBIE_REPEL_RADIUS_CELLS,
        )
        repel_x = 0.0
        repel_y = 0.0
        for cell, trapped_count in trapped_spiky_plant_counts.items():
            if trapped_count <= 0:
                continue
            hp = spiky_plants.get(cell)
            if hp is None or not hp.alive():
                continue
            dx = next_x - hp.x
            dy = next_y - hp.y
            dist = math.hypot(dx, dy)
            if dist > effect_radius:
                continue
            if dist <= 0.001:
                angle = RNG.uniform(0.0, math.tau)
                dx = math.cos(angle)
                dy = math.sin(angle)
                dist = 1.0
            falloff = 1.0 - (dist / effect_radius)
            count_mult = min(
                TRAPPED_ZOMBIE_REPEL_MAX_MULT,
                trapped_count * TRAPPED_ZOMBIE_REPEL_PER_STACK,
            )
            repel_mag = self.speed * count_mult * falloff
            repel_x += (dx / dist) * repel_mag
            repel_y += (dy / dist) * repel_mag
        return move_x + repel_x, move_y + repel_y

    def _apply_decay(self: Self) -> None:
        """Reduce zombie health over time and despawn when depleted."""
        self.vitals.apply_decay()

    def take_damage(
        self: Self,
        amount: int,
        *,
        source: str | None = None,
        now_ms: int,
    ) -> None:
        if amount <= 0 or not self.alive():
            return
        self.vitals.take_damage(amount, source=source, now_ms=now_ms)

    def _set_facing_bin(self: Self, new_bin: int) -> None:
        if new_bin == self.facing_bin:
            return
        center = self.rect.center
        self.facing_bin = new_bin
        self.image = self.directional_images[self.facing_bin]
        self.rect = self.image.get_rect(center=center)

    def _update_facing_from_movement(self: Self, dx: float, dy: float) -> None:
        new_bin = angle_bin_from_vector(dx, dy)
        if new_bin is None:
            return
        self._set_facing_bin(new_bin)

    def _apply_render_overlays(self: Self) -> None:
        base_surface = self.directional_images[self.facing_bin]
        self.image = apply_zombie_kind_overlay(
            base_surface=base_surface,
            kind=self.kind,
            facing_bin=self.facing_bin,
            collision_radius=self.collision_radius,
            wall_hug_side=self.wall_hug_side,
            wall_hug_last_side_has_wall=self.wall_hug_last_side_has_wall,
            lineformer_target_pos=self.lineformer_target_pos,
            zombie_pos=(self.x, self.y),
        )

    def _apply_paralyze_overlay(self: Self, now_ms: int) -> None:
        self._apply_render_overlays()
        self.image = self.image.copy()
        center = self.image.get_rect().center
        marker_size = max(6, int(self.radius * 0.8))
        draw_paralyze_marker_overlay(
            surface_out=self.image,
            now_ms=now_ms,
            blink_ms=PATROL_BOT_PARALYZE_BLINK_MS,
            center=center,
            size=marker_size,
            color=PATROL_BOT_PARALYZE_MARKER_COLOR,
            offset=int(self.radius * 0.4),
            width=2,
        )

    def _avoid_pitfalls(
        self: Self,
        pitfall_cells: set[tuple[int, int]],
        cell_size: int,
    ) -> tuple[float, float]:
        if cell_size <= 0 or not pitfall_cells:
            return 0.0, 0.0
        cell_x = int(self.x // cell_size)
        cell_y = int(self.y // cell_size)
        search_cells = 1
        avoid_radius = cell_size * 1.25
        max_strength = self.speed * 0.5
        push_x = 0.0
        push_y = 0.0
        for cy in range(cell_y - search_cells, cell_y + search_cells + 1):
            for cx in range(cell_x - search_cells, cell_x + search_cells + 1):
                if (cx, cy) not in pitfall_cells:
                    continue
                pit_x = (cx + 0.5) * cell_size
                pit_y = (cy + 0.5) * cell_size
                dx = self.x - pit_x
                dy = self.y - pit_y
                dist_sq = dx * dx + dy * dy
                if dist_sq <= 0:
                    continue
                dist = math.sqrt(dist_sq)
                if dist >= avoid_radius:
                    continue
                strength = (1.0 - dist / avoid_radius) * max_strength
                push_x += (dx / dist) * strength
                push_y += (dy / dist) * strength
        return push_x, push_y

    def update(
        self: Self,
        player_center: tuple[float, float],
        walls: list[Wall],
        nearby_zombies: Iterable[Zombie],
        electrified_cells: set[tuple[int, int]] | None = None,
        footprints: list[Footprint] | None = None,
        *,
        cell_size: int,
        layout: LevelLayout,
        now_ms: int,
        drift: tuple[float, float] = (0.0, 0.0),
        spiky_plants: dict[tuple[int, int], SpikyPlantLike] | None = None,
        trapped_spiky_plant_counts: dict[tuple[int, int], int] | None = None,
    ) -> None:
        if self.vitals.carbonized:
            self._apply_decay()
            return
        now = now_ms
        drift_x, drift_y = drift
        level_width = layout.field_rect.width
        level_height = layout.field_rect.height
        self._apply_decay()
        if not self.alive():
            return

        on_electrified_floor = False
        if cell_size > 0 and electrified_cells:
            current_cell = (int(self.x // cell_size), int(self.y // cell_size))
            on_electrified_floor = current_cell in electrified_cells
        if self.vitals.update_patrol_floor_paralyze(
            on_electrified_floor=on_electrified_floor,
            now_ms=now,
            paralyze_duration_ms=PATROL_BOT_PARALYZE_MS,
            damage_interval_frames=PATROL_BOT_ZOMBIE_DAMAGE_INTERVAL_FRAMES,
            damage_amount=PATROL_BOT_ZOMBIE_DAMAGE,
            apply_damage=lambda amount: self.take_damage(
                amount, source="patrol_bot", now_ms=now
            ),
        ):
            self.last_move_dx = 0.0
            self.last_move_dy = 0.0
            self._apply_paralyze_overlay(now)
            return
        dx_player = player_center[0] - self.x
        dy_player = player_center[1] - self.y
        dist_to_player_sq = dx_player * dx_player + dy_player * dy_player
        avoid_radius = max(SCREEN_WIDTH, SCREEN_HEIGHT) * 2
        avoid_radius_sq = avoid_radius * avoid_radius
        move_x, move_y = self.movement_strategy(
            self,
            cell_size,
            layout,
            player_center,
            nearby_zombies,
            footprints or [],
            now_ms=now,
        )
        move_x += drift_x
        move_y += drift_y

        # Puddle slow-down
        if is_in_puddle_cell(
            self.x,
            self.y,
            cell_size=cell_size,
            puddle_cells=layout.puddle_cells,
        ):
            move_x *= PUDDLE_SPEED_FACTOR
            move_y *= PUDDLE_SPEED_FACTOR
        if is_in_contaminated_cell(
            self.x,
            self.y,
            cell_size=cell_size,
            contaminated_cells=layout.zombie_contaminated_cells,
        ):
            move_x *= ZOMBIE_CONTAMINATED_SPEED_FACTOR
            move_y *= ZOMBIE_CONTAMINATED_SPEED_FACTOR
        move_x, move_y = self._slow_near_trapped_zombies(
            move_x, move_y, nearby_zombies
        )
        move_x, move_y = self._repel_from_loaded_spiky_plants(
            move_x,
            move_y,
            cell_size=cell_size,
            spiky_plants=spiky_plants,
            trapped_spiky_plant_counts=trapped_spiky_plant_counts,
        )

        if dist_to_player_sq <= avoid_radius_sq or self.kind == ZombieKind.WALL_HUGGER:
            move_x, move_y = self._avoid_other_zombies(move_x, move_y, nearby_zombies)
        move_x, move_y = apply_cell_edge_nudge(
            self.x,
            self.y,
            move_x,
            move_y,
            layout=layout,
            cell_size=cell_size,
        )
        self._update_facing_from_movement(move_x, move_y)
        self._apply_render_overlays()
        self.last_move_dx = move_x
        self.last_move_dy = move_y
        possible_walls = [
            w
            for w in walls
            if abs(w.rect.centerx - self.x) < 100 and abs(w.rect.centery - self.y) < 100
        ]

        final_x = self.x
        final_y = self.y
        if move_x:
            next_x = final_x + move_x
            for wall in possible_walls:
                if _circle_wall_collision(
                    (next_x, final_y), self.collision_radius, wall
                ):
                    if wall.alive():
                        wall._take_damage(amount=ZOMBIE_WALL_DAMAGE)
                    if wall.alive():
                        next_x = final_x
                        break
            final_x = next_x
        if move_y:
            next_y = final_y + move_y
            for wall in possible_walls:
                if _circle_wall_collision(
                    (final_x, next_y), self.collision_radius, wall
                ):
                    if wall.alive():
                        wall._take_damage(amount=ZOMBIE_WALL_DAMAGE)
                    if wall.alive():
                        next_y = final_y
                        break
            final_y = next_y

        if not (0 <= final_x < level_width and 0 <= final_y < level_height):
            self.kill()
            return

        self.x = final_x
        self.y = final_y
        self.rect.center = (int(self.x), int(self.y))

    def carbonize(self: Self) -> None:
        self.vitals.carbonize()


class TrappedZombie(pygame.sprite.Sprite):
    """A zombie or dog that has been trapped by a spiky plant."""

    def __init__(
        self: Self,
        x: float,
        y: float,
        kind: ZombieKind,
        health: int,
        max_health: int,
        facing_bin: int,
        radius: float,
        collision_radius: float,
        decay_duration_frames: float,
    ) -> None:
        super().__init__()
        self.x = x
        self.y = y
        self.kind = kind
        self.facing_bin = facing_bin
        self.radius = radius
        self.collision_radius = collision_radius
        self.shadow_radius = 0
        self.shadow_offset_scale = 1.0
        self.is_trapped = True

        if self.kind == ZombieKind.DOG:
            base_size = ZOMBIE_RADIUS * 2.0
            long_axis = base_size * ZOMBIE_DOG_LONG_AXIS_RATIO
            short_axis = base_size * ZOMBIE_DOG_SHORT_AXIS_RATIO
            self.directional_images = build_zombie_dog_directional_surfaces(
                long_axis, short_axis, is_trapped=True
            )
        else:
            self.directional_images = build_zombie_directional_surfaces(
                int(self.radius), draw_hands=False, is_trapped=True
            )

        self.image = self.directional_images[self.facing_bin]
        self.rect = self.image.get_rect(center=(int(x), int(y)))

        self.vitals = ZombieVitals(
            max_health=max_health,
            decay_duration_frames=decay_duration_frames,
            decay_min_speed_ratio=ZOMBIE_DECAY_MIN_SPEED_RATIO,
            carbonize_decay_frames=ZOMBIE_CARBONIZE_DECAY_FRAMES,
            on_health_ratio=lambda _r: None,  # No speed change needed
            on_kill=self.kill,
            on_carbonize=self._apply_carbonize_visuals,
        )
        self.vitals.health = health
        self.frame_counter = 0

    @property
    def health(self: Self) -> int:
        return self.vitals.health

    @property
    def max_health(self: Self) -> int:
        return self.vitals.max_health

    @property
    def carbonized(self: Self) -> bool:
        return self.vitals.carbonized

    def _apply_carbonize_visuals(self: Self) -> None:
        self.image = build_grayscale_image(self.image)

    def take_damage(
        self: Self,
        amount: int,
        *,
        source: str | None = None,
        now_ms: int = 0,
    ) -> None:
        if amount <= 0 or not self.alive():
            return
        self.vitals.take_damage(amount, source=source, now_ms=now_ms)

    def update(self: Self, *_args: Any, **_kwargs: Any) -> None:
        """Handle decay and jittering visuals."""
        self.vitals.apply_decay()
        if not self.alive():
            return

        # Jitter visuals at 1/4 speed
        if self.frame_counter % 4 == 0:
            ox = random.uniform(-1.0, 1.0)
            oy = random.uniform(-1.0, 1.0)
            self.rect.center = (int(self.x + ox), int(self.y + oy))
        self.frame_counter += 1

    def carbonize(self: Self) -> None:
        self.vitals.carbonize()



"""Gameplay-only constants."""

from __future__ import annotations

from ..entities_constants import ZOMBIE_DECAY_DURATION_FRAMES

# --- Survivor settings ---
SURVIVOR_SPEED_PENALTY_PER_PASSENGER = 0.08
SURVIVOR_OVERLOAD_DAMAGE_RATIO = 0.2
SURVIVOR_MESSAGE_DURATION_MS = 2000
INTRO_MESSAGE_DISPLAY_FRAMES = 180
SCREAM_MESSAGE_DISPLAY_FRAMES = 60

# --- Footprint settings (gameplay) ---
FOOTPRINT_STEP_DISTANCE = 40
FOOTPRINT_MAX = 320

# --- Zombie settings ---
MAX_ZOMBIES = 400
ZOMBIE_SPAWN_PLAYER_BUFFER = 230
ZOMBIE_TRACKER_DECAY_DURATION_FRAMES = ZOMBIE_DECAY_DURATION_FRAMES
FALLING_ZOMBIE_PRE_FX_MS = 600
FALLING_ZOMBIE_DURATION_MS = 450
FALLING_ZOMBIE_DUST_DURATION_MS = 220

# --- Car and fuel settings ---
FUEL_HINT_DURATION_MS = 1600

# --- Wall settings ---
OUTER_WALL_HEALTH = 9999999

# --- Render layers ---
LAYER_WALLS = 0
LAYER_ITEMS = 1
LAYER_HOUSEPLANTS = 1
LAYER_VEHICLES = 2
LAYER_ZOMBIES = 3
LAYER_PLAYERS = 4

__all__ = [
    "SURVIVOR_SPEED_PENALTY_PER_PASSENGER",
    "SURVIVOR_OVERLOAD_DAMAGE_RATIO",
    "SURVIVOR_MESSAGE_DURATION_MS",
    "INTRO_MESSAGE_DISPLAY_FRAMES",
    "SCREAM_MESSAGE_DISPLAY_FRAMES",
    "FOOTPRINT_STEP_DISTANCE",
    "FOOTPRINT_MAX",
    "MAX_ZOMBIES",
    "ZOMBIE_SPAWN_PLAYER_BUFFER",
    "ZOMBIE_TRACKER_DECAY_DURATION_FRAMES",
    "FALLING_ZOMBIE_PRE_FX_MS",
    "FALLING_ZOMBIE_DURATION_MS",
    "FALLING_ZOMBIE_DUST_DURATION_MS",
    "FUEL_HINT_DURATION_MS",
    "OUTER_WALL_HEALTH",
    "LAYER_WALLS",
    "LAYER_ITEMS",
    "LAYER_HOUSEPLANTS",
    "LAYER_VEHICLES",
    "LAYER_ZOMBIES",
    "LAYER_PLAYERS",
]



from __future__ import annotations

import math
from typing import TYPE_CHECKING

import pygame

try:
    from typing import Self
except ImportError:  # pragma: no cover - Python 3.10 fallback
    from typing_extensions import Self

from ..entities_constants import (
    PATROL_BOT_HUMANOID_PAUSE_MS,
    PATROL_BOT_COLLISION_RADIUS,
    PATROL_BOT_DIRECTION_COMMAND_RADIUS,
    PATROL_BOT_DIRECTION_COMMAND_HOLD_MS,
    PATROL_BOT_SPRITE_SIZE,
    PATROL_BOT_SPEED,
    PATROL_BOT_COLLISION_MARGIN,
    PUDDLE_SPEED_FACTOR,
)
from ..render_assets import angle_bin_from_vector, build_patrol_bot_directional_surfaces
from ..render_constants import ANGLE_BINS
from ..rng import get_rng
from ..surface_effects import is_in_puddle_cell
from ..world_grid import apply_cell_edge_nudge
from .movement import _circle_wall_collision, separate_circle_from_walls
from .walls import Wall

if TYPE_CHECKING:  # pragma: no cover - typing-only imports
    from .spiky_plant import SpikyPlant

RNG = get_rng()


class PatrolBot(pygame.sprite.Sprite):
    def __init__(self: Self, x: float, y: float) -> None:
        super().__init__()
        self.size = PATROL_BOT_SPRITE_SIZE
        self.radius = float(PATROL_BOT_COLLISION_RADIUS)
        self.facing_bin = 0
        self.collision_radius = float(self.radius)
        self.shadow_radius = max(1, int(self.collision_radius * 1.2))
        self.shadow_offset_scale = 1.0 / 3.0
        self.direction_command_radius = float(PATROL_BOT_DIRECTION_COMMAND_RADIUS)
        assert self.direction_command_radius < self.collision_radius
        self.directional_images_player = build_patrol_bot_directional_surfaces(
            self.size, arrow_scale=1.0
        )
        self.directional_images_auto = build_patrol_bot_directional_surfaces(
            self.size, arrow_scale=2.0 / 3.0
        )
        self.directional_images_waiting = build_patrol_bot_directional_surfaces(
            self.size, arrow_scale=1.0, marker_mode="diamond"
        )
        self.indicator_mode = "auto"
        self.directional_images = self.directional_images_auto
        self.image = self.directional_images[self.facing_bin]
        self.rect = self.image.get_rect(center=(int(x), int(y)))
        self.x = float(self.rect.centerx)
        self.y = float(self.rect.centery)
        self.speed = PATROL_BOT_SPEED
        self.direction = RNG.choice([(1, 0), (-1, 0), (0, 1), (0, -1)])
        self.turn_patterns = [
            [True, False],
            [True, True, False, False],
            [True, True, True, False, False, False],
            [True, True, True, True, False, False, False, False],
            [True, True, True, True, True, False, False, False, False, False],
        ]
        self.turn_pattern_idx = 0
        self.turn_step_idx = 0
        self.last_move_dx = 0.0
        self.last_move_dy = 0.0
        self.pause_until_ms = 0
        self.direction_command_hold_until_ms = 0
        self.awaiting_player_direction_input = False
        self._on_moving_floor_last = False

    def _set_facing_bin(self: Self, new_bin: int) -> None:
        if new_bin == self.facing_bin:
            return
        center = self.rect.center
        self.facing_bin = new_bin
        self.image = self.directional_images[self.facing_bin]
        self.rect = self.image.get_rect(center=center)

    def _set_indicator_mode(self: Self, mode: str) -> None:
        if mode == self.indicator_mode:
            return
        if mode not in ("auto", "player", "waiting"):
            mode = "auto"
        self.indicator_mode = mode
        center = self.rect.center
        if mode == "player":
            self.directional_images = self.directional_images_player
        elif mode == "waiting":
            self.directional_images = self.directional_images_waiting
        else:
            self.directional_images = self.directional_images_auto
        self.image = self.directional_images[self.facing_bin]
        self.rect = self.image.get_rect(center=center)

    def _update_facing_from_movement(self: Self, dx: float, dy: float) -> None:
        new_bin = angle_bin_from_vector(dx, dy)
        if new_bin is None:
            return
        self._set_facing_bin(new_bin)

    def _rotate_direction(self: Self, *, turn_right: bool) -> None:
        dx, dy = self.direction
        if turn_right:
            self.direction = (dy, -dx)
        else:
            self.direction = (-dy, dx)
        self._set_indicator_mode("auto")

    def _set_direction_from_player(
        self: Self, player: pygame.sprite.Sprite, *, now_ms: int
    ) -> None:
        input_bin = getattr(player, "input_facing_bin", None)
        if input_bin is None:
            return
        angle_rad = (input_bin % ANGLE_BINS) * (math.tau / ANGLE_BINS)
        dx = math.cos(angle_rad)
        dy = math.sin(angle_rad)
        if abs(dx) >= abs(dy):
            self.direction = (1, 0) if dx >= 0 else (-1, 0)
        else:
            self.direction = (0, 1) if dy >= 0 else (0, -1)
        self.direction_command_hold_until_ms = (
            now_ms + PATROL_BOT_DIRECTION_COMMAND_HOLD_MS
        )
        self._update_facing_from_movement(*self.direction)
        self._set_indicator_mode("player")

    def _handle_axis_collision(
        self: Self,
        *,
        next_x: float,
        next_y: float,
        walls: list[Wall],
        radius: float,
    ) -> tuple[float, float, bool]:
        final_x, final_y = next_x, next_y
        hit = False
        if next_x != self.x:
            for wall in walls:
                if _circle_wall_collision((next_x, self.y), radius, wall):
                    final_x = self.x
                    hit = True
                    break
        if next_y != self.y:
            for wall in walls:
                if _circle_wall_collision((final_x, next_y), radius, wall):
                    final_y = self.y
                    hit = True
                    break
        return final_x, final_y, hit

    def _resolve_circle_from_bot(
        self: Self,
        cx: float,
        cy: float,
        radius: float,
        other: "PatrolBot",
    ) -> tuple[float, float]:
        dx = cx - other.x
        dy = cy - other.y
        dist_sq = dx * dx + dy * dy
        min_dist = radius + other.collision_radius
        if dist_sq <= 0:
            return cx + min_dist, cy
        dist = dist_sq**0.5
        if dist >= min_dist:
            return cx, cy
        push = min_dist - dist
        return cx + (dx / dist) * push, cy + (dy / dist) * push

    def update(
        self: Self,
        walls: list[Wall],
        *,
        patrol_bot_group: pygame.sprite.Group | None = None,
        human_group: pygame.sprite.Group | None = None,
        player: pygame.sprite.Sprite | None = None,
        car: pygame.sprite.Sprite | None = None,
        parked_cars: list[pygame.sprite.Sprite] | None = None,
        cell_size: int,
        pitfall_cells: set[tuple[int, int]],
        fire_floor_cells: set[tuple[int, int]] | None = None,
        layout,
        drift: tuple[float, float] = (0.0, 0.0),
        now_ms: int,
        spiky_plants: dict[tuple[int, int], "SpikyPlant"] | None = None,
    ) -> None:
        now = now_ms
        drift_x, drift_y = drift

        if now < self.pause_until_ms:
            if player is not None and getattr(player, "alive", lambda: True)():
                hx, hy = player.rect.center
                hr = getattr(player, "collision_radius", None)
                if hr is None:
                    hr = max(player.rect.width, player.rect.height) / 2
                dx = self.x - hx
                dy = self.y - hy
                hit_range = (
                    self.collision_radius + PATROL_BOT_COLLISION_MARGIN
                ) + float(hr)
                if dx * dx + dy * dy <= hit_range * hit_range:
                    input_active = bool(getattr(player, "input_active", False))
                    if not input_active:
                        self.awaiting_player_direction_input = True
                        self._set_indicator_mode("waiting")
                    elif self.awaiting_player_direction_input:
                        self._set_direction_from_player(player, now_ms=now)
                        self.awaiting_player_direction_input = False
                    else:
                        self._set_indicator_mode("auto")
                else:
                    self.awaiting_player_direction_input = False
                    self._set_indicator_mode("auto")
            else:
                self.awaiting_player_direction_input = False
                self._set_indicator_mode("auto")
            self.last_move_dx = 0.0
            self.last_move_dy = 0.0
            return

        self.awaiting_player_direction_input = False

        cell_x = None
        cell_y = None
        if cell_size > 0:
            cell_x = int(self.rect.centerx // cell_size)
            cell_y = int(self.rect.centery // cell_size)
        on_floor = abs(drift_x) > 0.0 or abs(drift_y) > 0.0
        command_hold_active = now < self.direction_command_hold_until_ms
        forced_floor_dir = on_floor
        if on_floor and not command_hold_active:
            if abs(drift_x) >= abs(drift_y) and abs(drift_x) > 0.0:
                self.direction = (1, 0) if drift_x > 0 else (-1, 0)
            elif abs(drift_y) > 0.0:
                self.direction = (0, 1) if drift_y > 0 else (0, -1)
            self._set_indicator_mode("auto")

        move_x = float(self.direction[0]) * self.speed + drift_x
        move_y = float(self.direction[1]) * self.speed + drift_y

        # Puddle slow-down
        if is_in_puddle_cell(
            self.x,
            self.y,
            cell_size=cell_size,
            puddle_cells=layout.puddle_cells,
        ):
            move_x *= PUDDLE_SPEED_FACTOR
            move_y *= PUDDLE_SPEED_FACTOR

        move_x, move_y = apply_cell_edge_nudge(
            self.x,
            self.y,
            move_x,
            move_y,
            layout=layout,
            cell_size=cell_size,
        )
        self._update_facing_from_movement(
            float(self.direction[0]), float(self.direction[1])
        )
        self.last_move_dx = move_x
        self.last_move_dy = move_y

        next_x = self.x + move_x
        next_y = self.y + move_y
        collision_radius = self.collision_radius + PATROL_BOT_COLLISION_MARGIN
        final_x, final_y, hit_wall = self._handle_axis_collision(
            next_x=next_x,
            next_y=next_y,
            walls=walls,
            radius=collision_radius,
        )
        hit_bot = False
        possible_bots = []
        if patrol_bot_group:
            possible_bots = [
                b
                for b in patrol_bot_group
                if b is not self
                and b.alive()
                and abs(b.x - self.x) < 100
                and abs(b.y - self.y) < 100
            ]

        def _bot_collision(check_x: float, check_y: float) -> bool:
            for bot in possible_bots:
                dx = check_x - bot.x
                dy = check_y - bot.y
                hit_range = collision_radius + bot.collision_radius
                if dx * dx + dy * dy <= hit_range * hit_range:
                    return True
            return False

        if _bot_collision(final_x, final_y):
            hit_bot = True
            final_x = self.x
            final_y = self.y
            closest_bot = None
            closest_dist_sq = None
            for bot in possible_bots:
                dx = final_x - bot.x
                dy = final_y - bot.y
                dist_sq = dx * dx + dy * dy
                if closest_dist_sq is None or dist_sq < closest_dist_sq:
                    closest_bot = bot
                    closest_dist_sq = dist_sq

        hit_car = False
        car_candidates: list[pygame.sprite.Sprite] = []
        if car is not None and getattr(car, "alive", lambda: True)():
            car_candidates.append(car)
        if parked_cars:
            car_candidates.extend([c for c in parked_cars if c.alive()])
        for car_candidate in car_candidates:
            cx, cy = car_candidate.rect.center
            cr = getattr(car_candidate, "collision_radius", None)
            if cr is None:
                cr = max(car_candidate.rect.width, car_candidate.rect.height) / 2
            dx = final_x - cx
            dy = final_y - cy
            hit_range = collision_radius + float(cr)
            if dx * dx + dy * dy <= hit_range * hit_range:
                hit_car = True
                final_x = self.x
                final_y = self.y
                break

        hit_spiky_plant = False
        if spiky_plants and cell_size > 0:
            cx_idx = int(final_x // cell_size)
            cy_idx = int(final_y // cell_size)
            for dy in range(-1, 2):
                for dx in range(-1, 2):
                    hp = spiky_plants.get((cx_idx + dx, cy_idx + dy))
                    if hp and hp.alive():
                        h_dx = final_x - hp.x
                        h_dy = final_y - hp.y
                        h_hit_range = collision_radius + hp.collision_radius
                        if h_dx * h_dx + h_dy * h_dy <= h_hit_range * h_hit_range:
                            hit_spiky_plant = True
                            final_x = self.x
                            final_y = self.y
                            break
                if hit_spiky_plant:
                    break

        possible_humans = []
        if human_group:
            possible_humans.extend(
                [
                    h
                    for h in human_group
                    if h.alive()
                    and abs(h.rect.centerx - self.x) < 120
                    and abs(h.rect.centery - self.y) < 120
                ]
            )
        if player is not None and getattr(player, "alive", lambda: True)():
            if (
                abs(player.rect.centerx - self.x) < 120
                and abs(player.rect.centery - self.y) < 120
            ):
                possible_humans.append(player)

        def _humanoid_collision(check_x: float, check_y: float) -> bool:
            for human in possible_humans:
                hx, hy = human.rect.center
                hr = getattr(human, "collision_radius", None)
                if hr is None:
                    hr = max(human.rect.width, human.rect.height) / 2
                dx = check_x - hx
                dy = check_y - hy
                hit_range = collision_radius + float(hr)
                if dx * dx + dy * dy <= hit_range * hit_range:
                    return True
            return False

        hit_pitfall = False
        blocked_hazard_cells = set(pitfall_cells)
        if fire_floor_cells:
            blocked_hazard_cells.update(fire_floor_cells)
        if blocked_hazard_cells and cell_size > 0:
            lead_x = next_x + float(self.direction[0]) * collision_radius
            lead_y = next_y + float(self.direction[1]) * collision_radius
            cell_x = int(lead_x // cell_size)
            cell_y = int(lead_y // cell_size)
            if (cell_x, cell_y) in blocked_hazard_cells:
                hit_pitfall = True
                final_x = self.x
                final_y = self.y

        hit_outer = False
        if cell_size > 0:
            cell_x = int(final_x // cell_size)
            cell_y = int(final_y // cell_size)
            if (cell_x, cell_y) in layout.outer_wall_cells or (
                cell_x,
                cell_y,
            ) in layout.outside_cells:
                hit_outer = True
                final_x = self.x
                final_y = self.y
        if not (
            0 <= final_x < layout.field_rect.width
            and 0 <= final_y < layout.field_rect.height
        ):
            hit_outer = True
            final_x = self.x
            final_y = self.y

        if _humanoid_collision(final_x, final_y):
            self.pause_until_ms = now + PATROL_BOT_HUMANOID_PAUSE_MS
        elif (
            hit_wall
            or hit_pitfall
            or hit_bot
            or hit_car
            or hit_outer
            or hit_spiky_plant
        ):
            # Step back slightly to avoid corner lock, then rotate.
            backoff = max(0.5, self.speed * 0.5)
            final_x = self.x - float(self.direction[0]) * backoff
            final_y = self.y - float(self.direction[1]) * backoff
            final_x = min(layout.field_rect.width, max(0.0, final_x))
            final_y = min(layout.field_rect.height, max(0.0, final_y))
            if hit_bot and closest_bot is not None:
                final_x, final_y = self._resolve_circle_from_bot(
                    final_x, final_y, collision_radius, closest_bot
                )
            if hit_wall:
                (final_x, final_y), separated = separate_circle_from_walls(
                    (final_x, final_y),
                    collision_radius,
                    walls,
                    scale=1.0,
                    max_attempts=4,
                    first_extra_clearance=0.0,
                )
                final_x = min(layout.field_rect.width, max(0.0, final_x))
                final_y = min(layout.field_rect.height, max(0.0, final_y))
                if not separated:
                    final_x = self.x
                    final_y = self.y

            if not forced_floor_dir:
                # If we hit the outer boundary, reverse direction.
                if hit_outer:
                    self.direction = (-self.direction[0], -self.direction[1])
                    self._set_indicator_mode("auto")
                else:
                    pattern = self.turn_patterns[self.turn_pattern_idx]
                    preferred_turn = pattern[self.turn_step_idx]
                    self.turn_step_idx += 1
                    if self.turn_step_idx >= len(pattern):
                        self.turn_step_idx = 0
                        self.turn_pattern_idx = (self.turn_pattern_idx + 1) % len(
                            self.turn_patterns
                        )
                    self._rotate_direction(turn_right=preferred_turn)
                    # Keep the chosen turn even if it would remain blocked.

        level_width = layout.field_rect.width
        level_height = layout.field_rect.height
        if (
            final_x <= 0
            or final_y <= 0
            or final_x >= level_width
            or final_y >= level_height
        ):
            self.direction = (-self.direction[0], -self.direction[1])
            self._set_indicator_mode("auto")
            final_x = min(level_width - 1, max(1.0, final_x))
            final_y = min(level_height - 1, max(1.0, final_y))

        self.x = final_x
        self.y = final_y
        self.rect.center = (int(self.x), int(self.y))
        self._on_moving_floor_last = on_floor



from __future__ import annotations

from dataclasses import dataclass


def _clamp(value: float) -> int:
    return max(0, min(255, int(value)))


# Basic palette.
WHITE: tuple[int, int, int] = (255, 255, 255)
BLACK: tuple[int, int, int] = (0, 0, 0)
RED: tuple[int, int, int] = (255, 0, 0)
GREEN: tuple[int, int, int] = (0, 255, 0)
BLUE: tuple[int, int, int] = (0, 0, 255)
GRAY: tuple[int, int, int] = (100, 100, 100)
LIGHT_GRAY: tuple[int, int, int] = (200, 200, 200)
YELLOW: tuple[int, int, int] = (255, 255, 0)
ORANGE: tuple[int, int, int] = (255, 165, 0)
DARK_RED: tuple[int, int, int] = (139, 0, 0)


@dataclass(frozen=True)
class EnvironmentPalette:
    """Collection of colors that define the ambient environment."""

    floor_primary: tuple[int, int, int]
    floor_secondary: tuple[int, int, int]
    fall_zone_primary: tuple[int, int, int]
    fall_zone_secondary: tuple[int, int, int]
    outside: tuple[int, int, int]
    inner_wall: tuple[int, int, int]
    inner_wall_border: tuple[int, int, int]
    outer_wall: tuple[int, int, int]
    outer_wall_border: tuple[int, int, int]
    metal_floor_base: tuple[int, int, int]
    metal_floor_line: tuple[int, int, int]
    metal_floor_highlight: tuple[int, int, int]
    metal_floor_hairline: tuple[int, int, int]
    fire_floor_base: tuple[int, int, int]
    fire_grate_edge: tuple[int, int, int]
    fire_glass_base: tuple[int, int, int]


def _adjust_color(
    color: tuple[int, int, int], *, brightness: float = 1.0, saturation: float = 1.0
) -> tuple[int, int, int]:
    """Return color tinted by brightness/saturation multipliers."""

    r, g, b = color
    gray = 0.2126 * r + 0.7152 * g + 0.0722 * b

    def mix(component: int) -> int:
        value = gray + (component - gray) * saturation
        value *= brightness
        return _clamp(value)

    return mix(r), mix(g), mix(b)


_DEFAULT_AMBIENT_PALETTE_KEY = "default"
_NO_FLASHLIGHT_PALETTE_KEY = "no_flashlight"
DAWN_AMBIENT_PALETTE_KEY = "dawn"
STUDIO_AMBIENT_PALETTE_KEY = "studio"


# Base palette used throughout gameplay (matches the previous constants).
_DEFAULT_ENVIRONMENT_PALETTE = EnvironmentPalette(
    floor_primary=(43, 57, 70),
    floor_secondary=(50, 64, 79),
    fall_zone_primary=(84, 48, 29),
    fall_zone_secondary=(94, 54, 32),
    outside=(32, 60, 40),
    inner_wall=(125, 101, 78),
    inner_wall_border=(136, 110, 85),
    outer_wall=(136, 135, 128),
    outer_wall_border=(147, 146, 138),
    metal_floor_base=(52, 56, 62),
    metal_floor_line=(78, 84, 94),
    metal_floor_highlight=(104, 112, 126),
    metal_floor_hairline=(122, 130, 144),
    fire_floor_base=(88, 40, 22),
    fire_grate_edge=(58, 62, 70),
    fire_glass_base=(18, 22, 28),
)

# Dark, desaturated palette that sells the "alone without a flashlight" vibe.
_GLOOM_ENVIRONMENT_PALETTE = EnvironmentPalette(
    floor_primary=_adjust_color(
        _DEFAULT_ENVIRONMENT_PALETTE.floor_primary,
        brightness=0.8,
        saturation=0.75,
    ),
    floor_secondary=_adjust_color(
        _DEFAULT_ENVIRONMENT_PALETTE.floor_secondary,
        brightness=0.8,
        saturation=0.75,
    ),
    fall_zone_primary=_adjust_color(
        _DEFAULT_ENVIRONMENT_PALETTE.fall_zone_primary,
        brightness=0.8,
        saturation=0.75,
    ),
    fall_zone_secondary=_adjust_color(
        _DEFAULT_ENVIRONMENT_PALETTE.fall_zone_secondary,
        brightness=0.8,
        saturation=0.75,
    ),
    outside=_adjust_color(
        _DEFAULT_ENVIRONMENT_PALETTE.outside,
        brightness=0.8,
        saturation=0.75,
    ),
    inner_wall=_adjust_color(
        _DEFAULT_ENVIRONMENT_PALETTE.inner_wall,
        brightness=0.8,
        saturation=0.75,
    ),
    inner_wall_border=_adjust_color(
        _DEFAULT_ENVIRONMENT_PALETTE.inner_wall_border,
        brightness=0.8,
        saturation=0.75,
    ),
    outer_wall=_adjust_color(
        _DEFAULT_ENVIRONMENT_PALETTE.outer_wall,
        brightness=0.8,
        saturation=0.75,
    ),
    outer_wall_border=_adjust_color(
        _DEFAULT_ENVIRONMENT_PALETTE.outer_wall_border,
        brightness=0.8,
        saturation=0.75,
    ),
    metal_floor_base=_adjust_color(
        _DEFAULT_ENVIRONMENT_PALETTE.metal_floor_base,
        brightness=0.8,
        saturation=0.75,
    ),
    metal_floor_line=_adjust_color(
        _DEFAULT_ENVIRONMENT_PALETTE.metal_floor_line,
        brightness=0.8,
        saturation=0.75,
    ),
    metal_floor_highlight=_adjust_color(
        _DEFAULT_ENVIRONMENT_PALETTE.metal_floor_highlight,
        brightness=0.8,
        saturation=0.75,
    ),
    metal_floor_hairline=_adjust_color(
        _DEFAULT_ENVIRONMENT_PALETTE.metal_floor_hairline,
        brightness=0.8,
        saturation=0.75,
    ),
    fire_floor_base=_adjust_color(
        _DEFAULT_ENVIRONMENT_PALETTE.fire_floor_base,
        brightness=0.8,
        saturation=0.75,
    ),
    fire_grate_edge=_adjust_color(
        _DEFAULT_ENVIRONMENT_PALETTE.fire_grate_edge,
        brightness=0.8,
        saturation=0.75,
    ),
    fire_glass_base=_adjust_color(
        _DEFAULT_ENVIRONMENT_PALETTE.fire_glass_base,
        brightness=0.8,
        saturation=0.75,
    ),
)

_DAWN_ENVIRONMENT_PALETTE = EnvironmentPalette(
    floor_primary=(58, 70, 84),
    floor_secondary=(66, 78, 92),
    fall_zone_primary=(84, 39, 29),
    fall_zone_secondary=(95, 44, 33),
    outside=(118, 140, 104),
    inner_wall=(125, 101, 78),
    inner_wall_border=(136, 110, 85),
    outer_wall=(136, 135, 128),
    outer_wall_border=(147, 146, 138),
    metal_floor_base=_DEFAULT_ENVIRONMENT_PALETTE.metal_floor_base,
    metal_floor_line=_DEFAULT_ENVIRONMENT_PALETTE.metal_floor_line,
    metal_floor_highlight=_DEFAULT_ENVIRONMENT_PALETTE.metal_floor_highlight,
    metal_floor_hairline=_DEFAULT_ENVIRONMENT_PALETTE.metal_floor_hairline,
    fire_floor_base=_DEFAULT_ENVIRONMENT_PALETTE.fire_floor_base,
    fire_grate_edge=_DEFAULT_ENVIRONMENT_PALETTE.fire_grate_edge,
    fire_glass_base=_DEFAULT_ENVIRONMENT_PALETTE.fire_glass_base,
)

_STUDIO_ENVIRONMENT_PALETTE = EnvironmentPalette(
    floor_primary=WHITE,
    floor_secondary=WHITE,
    fall_zone_primary=_DEFAULT_ENVIRONMENT_PALETTE.fall_zone_primary,
    fall_zone_secondary=_DEFAULT_ENVIRONMENT_PALETTE.fall_zone_secondary,
    outside=WHITE,
    inner_wall=_DEFAULT_ENVIRONMENT_PALETTE.inner_wall,
    inner_wall_border=_DEFAULT_ENVIRONMENT_PALETTE.inner_wall_border,
    outer_wall=_DEFAULT_ENVIRONMENT_PALETTE.outer_wall,
    outer_wall_border=_DEFAULT_ENVIRONMENT_PALETTE.outer_wall_border,
    metal_floor_base=_DEFAULT_ENVIRONMENT_PALETTE.metal_floor_base,
    metal_floor_line=_DEFAULT_ENVIRONMENT_PALETTE.metal_floor_line,
    metal_floor_highlight=_DEFAULT_ENVIRONMENT_PALETTE.metal_floor_highlight,
    metal_floor_hairline=_DEFAULT_ENVIRONMENT_PALETTE.metal_floor_hairline,
    fire_floor_base=_DEFAULT_ENVIRONMENT_PALETTE.fire_floor_base,
    fire_grate_edge=_DEFAULT_ENVIRONMENT_PALETTE.fire_grate_edge,
    fire_glass_base=_DEFAULT_ENVIRONMENT_PALETTE.fire_glass_base,
)

_ENVIRONMENT_PALETTES: dict[str, EnvironmentPalette] = {
    _DEFAULT_AMBIENT_PALETTE_KEY: _DEFAULT_ENVIRONMENT_PALETTE,
    _NO_FLASHLIGHT_PALETTE_KEY: _GLOOM_ENVIRONMENT_PALETTE,
    DAWN_AMBIENT_PALETTE_KEY: _DAWN_ENVIRONMENT_PALETTE,
    STUDIO_AMBIENT_PALETTE_KEY: _STUDIO_ENVIRONMENT_PALETTE,
}


def get_environment_palette(key: str | None) -> EnvironmentPalette:
    """Return the color palette for the provided key (falls back to default)."""

    if not key:
        return _ENVIRONMENT_PALETTES[_DEFAULT_AMBIENT_PALETTE_KEY]
    return _ENVIRONMENT_PALETTES.get(
        key, _ENVIRONMENT_PALETTES[_DEFAULT_AMBIENT_PALETTE_KEY]
    )


def ambient_palette_key_for_flashlights(count: int) -> str:
    """Return the palette key for the provided flashlight inventory count."""

    return (
        _DEFAULT_AMBIENT_PALETTE_KEY
        if max(0, count) > 0
        else _NO_FLASHLIGHT_PALETTE_KEY
    )


# World colors (default palette versions preserved for backwards compatibility).
FOOTPRINT_COLOR: tuple[int, int, int] = (110, 200, 255)
STEEL_BEAM_COLOR: tuple[int, int, int] = (110, 50, 50)
STEEL_BEAM_LINE_COLOR: tuple[int, int, int] = (180, 90, 90)
FALL_ZONE_FLOOR_PRIMARY: tuple[int, int, int] = (
    _DEFAULT_ENVIRONMENT_PALETTE.fall_zone_primary
)
FALL_ZONE_FLOOR_SECONDARY: tuple[int, int, int] = (
    _DEFAULT_ENVIRONMENT_PALETTE.fall_zone_secondary
)


__all__ = [
    "WHITE",
    "BLACK",
    "RED",
    "GREEN",
    "BLUE",
    "GRAY",
    "LIGHT_GRAY",
    "YELLOW",
    "ORANGE",
    "DARK_RED",
    "DAWN_AMBIENT_PALETTE_KEY",
    "STUDIO_AMBIENT_PALETTE_KEY",
    "FOOTPRINT_COLOR",
    "STEEL_BEAM_COLOR",
    "STEEL_BEAM_LINE_COLOR",
    "FALL_ZONE_FLOOR_PRIMARY",
    "FALL_ZONE_FLOOR_SECONDARY",
    "EnvironmentPalette",
    "get_environment_palette",
    "ambient_palette_key_for_flashlights",
]
